# -*- encoding:utf-8 -*-
"""度量模块基础"""

import functools, logging
import numpy as np
import pandas as pd
import seaborn as sns
from ultron.factor.empyrical import stats
from ultron.kdutils import date as date_util
from ultron.kdutils import stats as stats_util
"""市场中1年交易日，默认250日"""
g_market_trade_year = 250


def valid_check(func):
    """检测度量的输入是否正常，非正常显示info，正常继续执行被装饰方法"""

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        if self.valid:
            return func(self, *args, **kwargs)
        else:
            logging.info('metrics input is invalid or zero order gen!')

    return wrapper


class Base(object):

    def __init__(self,
                 orders_pd,
                 action_pd,
                 capital,
                 benchmark,
                 enable_stocks_full_rate_factor=False):
        """
        :param orders_pd: 回测结果生成的交易订单构成的pd.DataFrame对象
        :param action_pd: 回测结果生成的交易行为构成的pd.DataFrame对象
        :param capital: 资金类UltronCapital实例化对象
        :param benchmark: 交易基准对象，UltronBenchmark实例对象
        :param enable_stocks_full_rate_factor: 是否开启满仓乘数
        """
        self.capital = capital
        self.orders_pd = orders_pd
        self.action_pd = action_pd
        self.benchmark = benchmark
        """
            满仓乘数，如果设置为True, 针对度量信息如收益等需要除self.stocks_full_rate_factor
        """
        self.enable_stocks_full_rate_factor = enable_stocks_full_rate_factor
        self.valid = False
        if self.orders_pd is not None and self.capital is not None and 'capital_blance' in self.capital.capital_pd:
            self.valid = True
        # ipython notebook下使用logging.info
        self.log_func = print  #logging.info if UltronEnv.g_is_ipython else print

    @valid_check
    def fit_metrics(self):
        """执行所有度量函数"""
        # TODO 根据ORDER数量大于一定阀值启动进度条
        # with UltronProgress(100, 0, label='metrics progress...') as pg:
        # pg.show(5)
        self._metrics_base_stats()
        # pg.show(50)
        self._metrics_sell_stats()
        # pg.show(80)
        self._metrics_action_stats()
        # pg.show(95)
        self._metrics_extend_stats()

    def fit_metrics_order(self):
        """对外接口，并非度量真实成交了的结果，只度量orders_pd，即不涉及资金的度量"""
        self._metrics_sell_stats()

    def _metrics_base_stats(self):
        """度量真实成交了的capital_pd，即涉及资金的度量"""
        # 平均资金利用率
        self.cash_utilization = 1 - (
            self.capital.capital_pd.cash_blance /
            self.capital.capital_pd.capital_blance).mean()

        # 默认不使用满仓乘数即stocks_full_rate_factor＝1
        self.stocks_full_rate_factor = 1
        if self.enable_stocks_full_rate_factor:
            # 计算满仓比例
            stocks_full_rate = (self.capital.capital_pd.stocks_blance /
                                self.capital.capital_pd.capital_blance)
            # 避免除0
            stocks_full_rate[stocks_full_rate == 0] = 1
            # 倒数得到满仓乘数
            self.stocks_full_rate_factor = (1 / stocks_full_rate)

        # 收益数据
        self.benchmark_returns = np.round(
            self.benchmark.kl_pd.close.pct_change(), 3)
        # 如果enable_stocks_full_rate_factor 则 * self.stocks_full_rate_factor的意义为随时都是满仓
        self.algorithm_returns = np.round(
            self.capital.capital_pd['capital_blance'].pct_change(),
            3) * self.stocks_full_rate_factor

        # 收益cum数据
        # noinspection PyTypeChecker
        self.algorithm_cum_returns = stats.cum_returns(self.algorithm_returns)
        self.benchmark_cum_returns = stats.cum_returns(self.benchmark_returns)

        # 最后一日的cum return
        self.benchmark_period_returns = self.benchmark_cum_returns[-1]
        self.algorithm_period_returns = self.algorithm_cum_returns[-1]

        # 交易天数
        self.num_trading_days = len(self.benchmark_returns)

        # 年化收益
        self.algorithm_annualized_returns = \
            (g_market_trade_year / self.num_trading_days) * self.algorithm_period_returns
        self.benchmark_annualized_returns = \
            (g_market_trade_year / self.num_trading_days) * self.benchmark_period_returns

        # 策略平均收益
        # noinspection PyUnresolvedReferences
        self.mean_algorithm_returns = self.algorithm_returns.cumsum(
        ) / np.arange(1, self.num_trading_days + 1, dtype=np.float64)
        # 波动率
        self.benchmark_volatility = stats.annual_volatility(
            self.benchmark_returns)
        # noinspection PyTypeChecker
        self.algorithm_volatility = stats.annual_volatility(
            self.algorithm_returns)

        # 夏普比率
        self.benchmark_sharpe = stats.sharpe_ratio(self.benchmark_returns)
        # noinspection PyTypeChecker
        self.algorithm_sharpe = stats.sharpe_ratio(self.algorithm_returns)

        # 信息比率
        # noinspection PyUnresolvedReferences
        self.information = stats.information_ratio(
            self.algorithm_returns.values, self.benchmark_returns.values)

        # 阿尔法, 贝塔
        # noinspection PyUnresolvedReferences
        self.alpha, self.beta = stats.alpha_beta_aligned(
            self.algorithm_returns.values, self.benchmark_returns.values)

        # 最大回撤
        # noinspection PyUnresolvedReferences
        self.max_drawdown = stats.max_drawdown(self.algorithm_returns.values)

    def _metrics_sell_stats(self):
        """并非度量真实成交了的结果，只度量orders_pd，即认为没有仓位管理和资金量限制前提下的表现"""
        # 根据order中的数据，计算盈利比例
        self.orders_pd['profit_cg'] = self.orders_pd['profit'] / (
            self.orders_pd['buy_price'] * self.orders_pd['buy_cnt'])
        # 为了显示方便及明显
        self.orders_pd['profit_cg_hunder'] = self.orders_pd['profit_cg'] * 100
        # 成交了的pd isin win or loss
        deal_pd = self.orders_pd[self.orders_pd['sell_type'].isin(
            ['win', 'loss'])]
        # 卖出原因get_dummies进行离散化
        dumm_sell = pd.get_dummies(deal_pd.sell_type_extra)
        dumm_sell_t = dumm_sell.T
        # 为plot_sell_factors函数生成卖出生效因子分布
        self.dumm_sell_t_sum = dumm_sell_t.sum(axis=1)

        # 买入因子唯一名称get_dummies进行离散化
        dumm_buy = pd.get_dummies(deal_pd.buy_factor)
        dumm_buy = dumm_buy.T
        # 为plot_buy_factors函数生成卖出生效因子分布
        self.dumm_buy_t_sum = dumm_buy.sum(axis=1)

        self.orders_pd['buy_date'] = self.orders_pd['buy_date'].astype(int)
        self.orders_pd[self.orders_pd['result'] != 0]['sell_date'].astype(
            int, copy=False)
        # 因子的单子的持股时间长度计算
        self.orders_pd['keep_days'] = self.orders_pd.apply(
            lambda x: date_util.diff(
                x['buy_date'],
                date_util.current_date_int()
                if x['result'] == 0 else x['sell_date']),
            axis=1)
        # 筛出已经成交了的单子
        self.order_has_ret = self.orders_pd[self.orders_pd['result'] != 0]

        # 筛出未成交的单子
        self.order_keep = self.orders_pd[self.orders_pd['result'] == 0]

        xt = self.order_has_ret.result.value_counts()
        # 计算胜率
        if xt.shape[0] == 2:
            win_rate = xt[1] / xt.sum()
        elif xt.shape[0] == 1:
            win_rate = xt.index[0]
        else:
            win_rate = 0
        self.win_rate = win_rate
        # 策略持股天数平均值
        self.keep_days_mean = self.orders_pd['keep_days'].mean()
        # 策略持股天数中位数
        self.keep_days_median = self.orders_pd['keep_days'].median()

        # 策略期望收益
        self.gains_mean = self.order_has_ret[
            self.order_has_ret['profit_cg'] > 0].profit_cg.mean()
        if np.isnan(self.gains_mean):
            self.gains_mean = 0.0
        # 策略期望亏损
        self.losses_mean = self.order_has_ret[
            self.order_has_ret['profit_cg'] < 0].profit_cg.mean()
        if np.isnan(self.losses_mean):
            self.losses_mean = 0.0

        # 忽略仓位控的前提下，即假设每一笔交易使用相同的资金，策略的总获利交易获利比例和
        profit_cg_win_sum = self.order_has_ret[
            self.order_has_ret['profit_cg'] > 0].profit.sum()
        # 忽略仓位控的前提下，即假设每一笔交易使用相同的资金，策略的总亏损交易亏损比例和
        profit_cg_loss_sum = self.order_has_ret[
            self.order_has_ret['profit_cg'] < 0].profit.sum()

        if profit_cg_win_sum * profit_cg_loss_sum == 0 and profit_cg_win_sum + profit_cg_loss_sum > 0:
            # 其中有一个是0的，要转换成一个最小统计单位计算盈亏比，否则不需要
            if profit_cg_win_sum == 0:
                profit_cg_win_sum = 0.01
            if profit_cg_loss_sum == 0:
                profit_cg_win_sum = 0.01

        #  忽略仓位控的前提下，计算盈亏比
        self.win_loss_profit_rate = 0 if profit_cg_loss_sum == 0 else -round(
            profit_cg_win_sum / profit_cg_loss_sum, 4)
        #  忽略仓位控的前提下，计算所有交易单的盈亏总会
        self.all_profit = self.order_has_ret['profit'].sum()

    def _metrics_action_stats(self):
        """度量真实成交了的action_pd 计算买入资金的分布平均性，及是否有良好的分布"""
        action_pd = self.action_pd
        # 只选生效的, 由于忽略非交易日, 大概有多出0.6的误差
        self.act_buy = action_pd[action_pd.action.isin(['buy'])
                                 & action_pd.deal.isin([True])]
        # drop重复的日期上的行为，只保留一个，cp_date形如下所示
        cp_date = self.act_buy['Date'].drop_duplicates()
        """
            cp_date
            0      20141024
            2      20141029
            20     20150127
            21     20150205
            23     20150213
            25     20150218
            31     20150310
            34     20150401
            36     20150409
            39     20150422
            41     20150423
            44     20150428
            58     20150609
            59     20150610
            63     20150624
            66     20150715
            67     20150717
        """
        dt_fmt = cp_date.apply(
            lambda order: date_util.str_to_datetime(str(order), '%Y%m%d'))
        dt_fmt = dt_fmt.apply(lambda order: (order - dt_fmt.iloc[0]).days)
        # 前后两两生效交易时间相减
        self.diff_dt = dt_fmt - dt_fmt.shift(1)
        # 计算平均生效间隔时间
        self.effect_mean_day = self.diff_dt.mean()

        if self.act_buy.empty:
            self.act_buy['cost'] = 0
            self.cost_stats = 0
            self.buy_deal_rate = 0
        else:
            self.act_buy['cost'] = self.act_buy.apply(
                lambda order: order.Price * order.Cnt, axis=1)
            # 计算cost各种统计度量值
            self.cost_stats = stats_util.stats_namedtuple(self.act_buy['cost'])

            buy_action_pd = action_pd[action_pd['action'] == 'buy']
            buy_action_pd_deal = buy_action_pd['deal']
            # 计算资金对应的成交比例
            self.buy_deal_rate = buy_action_pd_deal.sum(
            ) / buy_action_pd_deal.count()

    def _metrics_extend_stats(self):
        """子类可扩展的metrics方法，子类在此方法中可定义自己需要度量的值"""
        pass


import numpy as np
import pandas as pd

def factor_coverage_seq(factors_data, factor_name):
    coverage = factors_data.sort_values(by=['trade_date', 'code']).groupby(
        ['trade_date']).apply(
        lambda x: 1 - np.isnan(x[factor_name].values).sum() / len(x))
    coverage.name = factor_name
    return coverage


def code_coverage_seq(factors_data, factors, threshold):
    res = []
    for col in factors:
        rts = factors_data.sort_values(by=['trade_date']).groupby(
            ['trade_date']).apply(
                lambda x: 1 - np.isnan(
                    x[col].values).sum() / len(x))
        res.append(rts)
    coverage = pd.concat(res,axis=1)
    valid_col = (coverage >= threshold).all()
    valid_col = valid_col[valid_col].index
    return len(valid_col)/len(factors)


# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: cache_data.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='cache_data.proto',
  package='ultron.cache_data',
  syntax='proto2',
  serialized_options=None,
  serialized_pb=_b('\n\x10\x63\x61\x63he_data.proto\x12\x11ultron.cache_data\">\n\tCacheData\x12\x0f\n\x07session\x18\x01 \x02(\x0c\x12\x0b\n\x03key\x18\x02 \x02(\x0c\x12\x13\n\x0b\x64\x61ta_stream\x18\x03 \x02(\x0c')
)




_CACHEDATA = _descriptor.Descriptor(
  name='CacheData',
  full_name='ultron.cache_data.CacheData',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='session', full_name='ultron.cache_data.CacheData.session', index=0,
      number=1, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='key', full_name='ultron.cache_data.CacheData.key', index=1,
      number=2, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='data_stream', full_name='ultron.cache_data.CacheData.data_stream', index=2,
      number=3, type=12, cpp_type=9, label=2,
      has_default_value=False, default_value=_b(""),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto2',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=39,
  serialized_end=101,
)

DESCRIPTOR.message_types_by_name['CacheData'] = _CACHEDATA
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

CacheData = _reflection.GeneratedProtocolMessageType('CacheData', (_message.Message,), dict(
  DESCRIPTOR = _CACHEDATA,
  __module__ = 'cache_data_pb2'
  # @@protoc_insertion_point(class_scope:ultron.cache_data.CacheData)
  ))
_sym_db.RegisterMessage(CacheData)


# @@protoc_insertion_point(module_scope)



# -*- encoding:utf-8 -*-
"""
    相关系数具体计算功能实现模块
"""

import numpy as np
import pandas as pd
import scipy.stats as stats

from enum import Enum
from ultron.kdutils.decorator import consume_time
from ultron.ump.core.fixes import rankdata
from ultron.ump.core.helper import pd_rolling_corr


class ECoreCorrType(Enum):
    """
        ECoreCorrType: 相关系数计算方法
    """
    """皮尔逊相关系数计算"""
    E_CORE_TYPE_PEARS = 'pears'
    """斯皮尔曼相关系数计算"""
    E_CORE_TYPE_SPERM = 'sperm'
    """基于PEARS使用序列＋－符号相关系数"""
    E_CORE_TYPE_SIGN = 'sign'
    """基于PEARS使用移动时间加权相关系数"""
    E_CORE_TYPE_ROLLING = 'rolling'

    @classmethod
    def task_cnt(cls):
        """ECoreCorrType暂时支持的相关计算方法个数"""
        return 4


"""加权移动相关系数计算默认使用60d"""
g_rolling_corr_window = 60


def spearmanr(a, b=None, axis=0, p_value=False):
    """
    如果需要计算p_value使用stats.spearmanr计算，否则使用rankdata配合使用np.apply_along_axis，
    进行spearmanr相关计算，因为计算p_value耗时
    :param a: 可迭代序列a
    :param b: 可迭代序列b
    :param axis: 系数计算作用轴方向
    :param p_value 是否需要计算p_value
    :return: p_value 是True是返回 scipy.cost_stats.SpearmanrResult
             p_value 是False 返回 np.array 的二维方阵
    """
    if p_value:
        # 需要计算p_value使用stats.spearmanr计算
        return stats.spearmanr(a=a, b=b, axis=axis)
    else:
        # 使用rankdata配合使用np.apply_along_axis
        a, outaxis = _chk_asarray(a, axis)
        ar = np.apply_along_axis(rankdata, outaxis, a)
        br = None
        if b is not None:
            b, axisout = _chk_asarray(b, axis)
            br = np.apply_along_axis(rankdata, axisout, b)
        # 返回 np.array 的二维方阵
        return np.corrcoef(ar, br, rowvar=outaxis)


def corr_xy(x, y, similar_type=ECoreCorrType.E_CORE_TYPE_PEARS, **kwargs):
    """
    计算两个可迭代序列相关系数对外函数
    :param x: 可迭代序列
    :param y: 可迭代序列
    :param similar_type: ECoreCorrType, 默认值ECoreCorrType.E_CORE_TYPE_PEARS
    :return: x与y的相关系数返回值
    """
    if similar_type == ECoreCorrType.E_CORE_TYPE_SIGN:
        # 序列＋－符号相关系数, 使用np.sign取符号后，再np.corrcoef计算
        x = np.sign(x)
        y = np.sign(y)
        similar_type = ECoreCorrType.E_CORE_TYPE_PEARS

    # noinspection PyTypeChecker
    if np.all(x == x[0]) or np.all(y == y[0]):
        # 如果全序列唯一不能使用相关计算，使用相同的数和与总数的比例
        # noinspection PyUnresolvedReferences
        return (x == y).sum() / x.count()

    if similar_type == ECoreCorrType.E_CORE_TYPE_PEARS:
        # 皮尔逊相关系数计算
        return np.corrcoef(x, y)[0][1]
    elif similar_type == ECoreCorrType.E_CORE_TYPE_SPERM:
        # 斯皮尔曼相关系数计算, 使用自定义spearmanr，不计算p_value
        return spearmanr(x, y)[0][1]
    elif similar_type == ECoreCorrType.E_CORE_TYPE_ROLLING:
        # pop参数window，默认使用g_rolling_corr_window
        window = kwargs.pop('window', g_rolling_corr_window)

        # 加权时间需要可迭代序列是pd.Series
        if not isinstance(x, pd.Series):
            x = pd.Series(x)
        if not isinstance(y, pd.Series):
            y = pd.Series(y)
        return rolling_corr(x, y, window=window)


# noinspection PyUnresolvedReferences
def corr_matrix(df, similar_type=ECoreCorrType.E_CORE_TYPE_PEARS, **kwargs):
    """
    与corr_xy的区别主要是，非两两corr计算，输入参数除类别外，只有一个矩阵的输入，且输入必须为pd.DataFrame对象 or np.array
    :param df: pd.DataFrame or np.array, 之所以叫df，是因为在内部会统一转换为pd.DataFrame
    :param similar_type: ECoreCorrType, 默认值ECoreCorrType.E_CORE_TYPE_PEARS
    :return: pd.DataFrame对象
    """
    if isinstance(df, np.ndarray):
        # 把np.ndarray转DataFrame，便统一处理
        df = pd.DataFrame(df)

    if not isinstance(df, pd.DataFrame):
        raise TypeError('df must pd.DataFrame object!!!')

    # FIXME 这里不应该支持ECoreCorrType.E_CORE_TYPE_PEARS.value，只严格按照ECoreCorrType对象相等
    if similar_type == ECoreCorrType.E_CORE_TYPE_PEARS or similar_type == ECoreCorrType.E_CORE_TYPE_PEARS.value:
        # 皮尔逊相关系数计算
        corr = np.corrcoef(df.T)
    elif similar_type == ECoreCorrType.E_CORE_TYPE_SPERM or similar_type == ECoreCorrType.E_CORE_TYPE_SPERM.value:
        # 斯皮尔曼相关系数计算, 使用自定义spearmanr，不计算p_value
        corr = spearmanr(df)
    elif similar_type == ECoreCorrType.E_CORE_TYPE_SIGN or similar_type == ECoreCorrType.E_CORE_TYPE_SIGN.value:
        # 序列＋－符号相关系数, 使用np.sign取符号后，再np.corrcoef计算
        corr = np.corrcoef(np.sign(df.T))
    elif similar_type == ECoreCorrType.E_CORE_TYPE_ROLLING or similar_type == ECoreCorrType.E_CORE_TYPE_ROLLING.value:
        # pop参数window，默认使用g_rolling_corr_window
        window = kwargs.pop('window', g_rolling_corr_window)
        corr = rolling_corr(df, window=window)
    else:
        # 还是给个默认的corr计算np.corrcoef(df.T)
        corr = np.corrcoef(df.T)
    # 将计算结果的corr转换为pd.DataFrame对象，行和列索引都使用df.columns
    corr = pd.DataFrame(corr, index=df.columns, columns=df.columns)
    return corr


@consume_time
def rolling_corr(df, ss=None, window=g_rolling_corr_window):
    """
    滑动窗口按时间权重计算相关系数

    :param df: pd.DataFrame对象或者pd.Series对象
    e.g.
                usA	    usAA	usAAC
    2015/7/27	0.76	-1.94	0.59
    2015/7/28	2.12	2.6	    1.3
    2015/7/29	-0.12	2.94	-1.34
    2015/7/30	1.41	-1.77	-4.04
    2015/7/31	-0.05	-1.1	1.39
    ......

    :param ss: pd.Series对象, ss的大小需要与df.shape[0]一样
    e.g.
                usA
    2015/7/27	0.76
    2015/7/28	2.12
    2015/7/29	-0.12
    2015/7/30	1.41
    2015/7/31	-0.05
    ......
    :param window: 窗口大小, 默认值g_rolling_corr_window，即60d
    :return: 当ss为None时返回df.shape[1]大小的相关系数二维方阵，否则，返回长度为df.shape[1]的相关系数一维数组
    """
    if window < 1 or window > df.shape[0]:
        raise TypeError('window out of index, must in [{},{}]'.format(
            1, df.shape[0]))
    rolling_window = list(
        zip(np.arange(df.shape[0] - window + 1),
            np.arange(window, df.shape[0] + 1)))
    """
        通过list(zip(np.arange(df.shape[0] - window + 1), np.arange(window, df.shape[0] + 1)))
        生成rolling_window，rolling_window形如下：
        [(0, 60), (1, 61), (2, 62), (3, 63), (4, 64), (5, 65), (6, 66), (7, 67), (8, 68),
        (9, 69), (10, 70), (11, 71), (12, 72), (13, 73), (14, 74), (15, 75), (16, 76),
        (17, 77), (18, 78), (19, 79), (20, 80), (21, 81), (22, 82), (23, 83), (24, 84),
         (25, 85), (26, 86), (27, 87), (28, 88),........]
    """
    weights = np.linspace(0, 1, len(rolling_window))
    # 随着时间的推移，过去的时间占得比重越来越少
    # noinspection PyUnresolvedReferences
    weights = weights / weights.sum()
    """
        weights即权重序列形如下所示：
        [0.      0.      0.      0.      0.      0.0001  0.0001  0.0001  0.0001
        0.0001  0.0001  0.0001  0.0001.........................................
        0.0044  0.0044  0.0044  0.0044  0.0044  0.0044  0.0044  0.0044  0.0045
        ........................................0.0045  0.0045  0.0045  0.0045]
    """
    corr = 0
    if ss is None:
        # 如果不使用pd_rolling_corr，需要copy，保证df == np.inf的会修改
        df = df.copy()
        df[df == np.inf] = 0
        # 迭代rolling_window下的时间窗口，使用np.corrcoef，比使用pd_rolling_corr效果高很多
        for (s, e) in rolling_window:
            # eg. rolling_window第一个即为np.corrcoef(df.iloc[0:60].T)
            window_corr = np.corrcoef(df.iloc[s:e].T)
            window_corr[np.isinf(window_corr) | np.isnan(window_corr)] = 0
            # 当前窗口下的相关系数乘以权重, window_corr * weights[s]为df.shape[1]大小的相关系数二维方阵
            corr += window_corr * weights[s]
    else:
        # 针对两个输入序列使用pd_rolling_corr
        window_corr = pd_rolling_corr(df, ss, window=window)
        window_corr.dropna(inplace=True, how='all')
        """
            pd_rolling_corr返回的window_corr形如, 即每一个子窗口corr:

            59     0.234309
            60     0.237821
            61     0.244905
            62     0.242731
            63     0.249227
            ...............
            499    0.154148
            500    0.138837
            501    0.133806
            502    0.135788
            503    0.138762
        """
        window_corr[window_corr == np.inf] = 0
        for ind in np.arange(window_corr.shape[0]):
            # 对应天的相关系数 ＊ 对应天的系统权重，长度为df.shape[1]的相关系数一维数组
            corr += window_corr.iloc[ind] * weights[ind]
    return corr


def _chk_asarray(a, axis):
    """内部函数，为spearmanr下不需要计算p_value的情况下，为apply_along_axis转换数据"""
    if axis is None:
        a = np.ravel(a)
        outaxis = 0
    else:
        a = np.asarray(a)
        outaxis = axis

    if a.ndim == 0:
        a = np.atleast_1d(a)

    return a, outaxis


# -*- coding: utf-8 -*-

from ultron.optimize.model.linearmodel import ConstLinearModel
from ultron.optimize.model.linearmodel import LassoRegression
from ultron.optimize.model.linearmodel import BayesianRegression
from ultron.optimize.model.linearmodel import RidgeRegression
from ultron.optimize.model.linearmodel import TweedieRegression
from ultron.optimize.model.linearmodel import HuberRegression
from ultron.optimize.model.linearmodel import SGDRegression
from ultron.optimize.model.linearmodel import PassiveAggressiveRegression
from ultron.optimize.model.linearmodel import TheilSenRegression
from ultron.optimize.model.linearmodel import LinearRegression
from ultron.optimize.model.linearmodel import LogisticRegression
from ultron.optimize.model.treemodel import RandomForestClassifier
from ultron.optimize.model.treemodel import RandomForestRegressor
from ultron.optimize.model.treemodel import ExtraTreesClassifier
from ultron.optimize.model.treemodel import ExtraTreesRegressor
from ultron.optimize.model.treemodel import BaggingClassifier
from ultron.optimize.model.treemodel import BaggingRegressor
from ultron.optimize.model.treemodel import AdaBoostClassifier
from ultron.optimize.model.treemodel import AdaBoostRegressor
from ultron.optimize.model.treemodel import GradientBoostingClassifier
from ultron.optimize.model.treemodel import GradientBoostingRegressor
from ultron.optimize.model.treemodel import DecisionTreeClassifier
from ultron.optimize.model.treemodel import DecisionTreeRegressor
from ultron.optimize.model.treemodel import XGBClassifier
from ultron.optimize.model.treemodel import XGBRegressor
from ultron.optimize.model.treemodel import XGBTrainer
from ultron.optimize.model.treemodel import LGBMClassifier
from ultron.optimize.model.treemodel import LGBMRegressor
from ultron.optimize.model.treemodel import LGBMTrainer
from ultron.optimize.model.treemodel import VotingRegressor
from ultron.optimize.model.treemodel import VotingClassifier
from ultron.optimize.model.treemodel import StackingClassifier
from ultron.optimize.model.treemodel import StackingRegressor
from ultron.optimize.model.modelbase import ModelBase


def load_model(model_desc):
    model_name = model_desc['model_name']
    model_name_parts = set(model_name.split('.'))

    if 'ConstLinearModel' in model_name_parts:
        return ConstLinearModel.load(model_desc)
    elif 'LinearRegression' in model_name_parts:
        return LinearRegression.load(model_desc)
    elif 'LassoRegression' in model_name_parts:
        return LassoRegression.load(model_desc)
    elif 'BayesianRegression' in model_name_parts:
        return BayesianRegression.load(model_desc)
    elif 'RidgeRegression' in model_name_parts:
        return RidgeRegression.load(model_desc)
    elif 'TweedieRegression' in model_name_parts:
        return TweedieRegression.load(model_desc)
    elif 'HuberRegression' in model_name_parts:
        return HuberRegression.load(model_desc)
    elif 'SGDRegression' in model_name_parts:
        return SGDRegression.load(model_desc)
    elif 'PassiveAggressiveRegression' in model_name_parts:
        return PassiveAggressiveRegression.load(model_desc)
    elif 'TheilSenRegression' in model_name_parts:
        return TheilSenRegression.load(model_desc)
    elif 'LogisticRegression' in model_name_parts:
        return LogisticRegression.load(model_desc)
    elif 'RandomForestRegressor' in model_name_parts:
        return RandomForestRegressor.load(model_desc)
    elif 'RandomForestClassifier' in model_name_parts:
        return RandomForestClassifier.load(model_desc)
    elif 'ExtraTreesClassifier' in model_name_parts:
        return ExtraTreesClassifier.load(model_desc)
    elif 'ExtraTreesRegressor' in model_name_parts:
        return ExtraTreesRegressor.load(model_desc)
    elif 'BaggingClassifier' in model_name_parts:
        return BaggingClassifier.load(model_desc)
    elif 'BaggingRegressor' in model_name_parts:
        return BaggingRegressor.load(model_desc)
    elif 'AdaBoostClassifier' in model_name_parts:
        return AdaBoostClassifier.load(model_desc)
    elif 'AdaBoostRegressor' in model_name_parts:
        return AdaBoostRegressor.load(model_desc)
    elif 'GradientBoostingRegressor' in model_name_parts:
        return GradientBoostingRegressor.load(model_desc)
    elif 'GradientBoostingClassifier' in model_name_parts:
        return GradientBoostingClassifier.load(model_desc)
    elif 'DecisionTreeClassifier' in model_name_parts:
        return DecisionTreeClassifier.load(model_desc)
    elif 'DecisionTreeRegressor' in model_name_parts:
        return DecisionTreeRegressor.load(model_desc)
    elif 'XGBRegressor' in model_name_parts:
        return XGBRegressor.load(model_desc)
    elif 'XGBClassifier' in model_name_parts:
        return XGBClassifier.load(model_desc)
    elif 'XGBTrainer' in model_name_parts:
        return XGBTrainer.load(model_desc)
    elif 'LGBMRegressor' in model_name_parts:
        return LGBMRegressor.load(model_desc)
    elif 'LGBMClassifier' in model_name_parts:
        return LGBMClassifier.load(model_desc)
    elif 'LGBMTrainer' in model_name_parts:
        return LGBMTrainer.load(model_desc)
    elif 'VotingRegressor' in model_name_parts:
        return VotingRegressor.load(model_desc)
    elif 'VotingClassifier' in model_name_parts:
        return VotingClassifier.load(model_desc)
    elif 'StackingClassifier' in model_name_parts:
        return StackingClassifier.load(model_desc)
    elif 'StackingRegressor' in model_name_parts:
        return StackingRegressor.load(model_desc)
    else:
        raise ValueError(
            '{0} is not currently supported in model loader.'.format(
                model_name))



# -*- coding: utf-8 -*-
import numpy as np
import datetime, time, hashlib, json, copy
from ultron.optimize.model.modelbase import load_module
from ultron.optimize.geneticist.fitness import model_fitness
from ultron.utilities.utils import NumpyEncoder

import warnings

warnings.filterwarnings("ignore")


class Program(object):

    def __init__(self,
                 params_sets,
                 model_sets,
                 method,
                 random_state,
                 p_point_replace,
                 gen,
                 fitness,
                 model_name=None,
                 params=None,
                 program=None,
                 parents=None):
        self._id = str(
            int(time.time() * 1000000 + datetime.datetime.now().microsecond))
        self._init_method = method
        self._random_state = random_state
        self._p_point_replace = p_point_replace
        self._gen = gen
        self._fitness = fitness
        self._program = program
        self._parents = parents
        self._raw_fitness = None
        self._create_time = datetime.datetime.now()
        if self._program is None:
            self._program = self.build_program(random_state, params_sets,
                                               model_sets, model_name)
        else:
            self._program = self.reset_program(program)
        self._params = self._program['params']
        self._model_name = self._program['model_name']
        self._params_sets = params_sets[self._model_name]
        self._name = self._model_name + '_' + self._id
        self.create_identification()
        self._is_valid = True

    def reset_program(self, program):
        return program

    def build_program(self, random_state, params_sets, model_sets, model_name):
        ### 选择模型
        if model_name is None:
            model_name = random_state.randint(len(model_sets))
            model_name = model_sets[model_name]
        ##获取模型对应参数集
        model_params_sets = params_sets[model_name]
        params = {}
        for key in model_params_sets.keys():
            params[key] = random_state.choice(model_params_sets[key])
        #self._model = load_module(self._model_name)(**params)
        return {'model_name': model_name, 'params': params}

    def log(self):
        print("name:{0},gen:{1},params:{2},fitness:{3},method:{4},token:{5}".
              format(self._name, self._gen, self._params, self._raw_fitness,
                     self._init_method, self._identification))

    def create_identification(self):
        m = hashlib.md5()
        try:
            token = self.transform()
        except Exception as e:
            #ID为key
            token = self._name
        if token is None:
            token = self._name
        m.update(bytes(token, encoding='UTF-8'))
        self._identification = m.hexdigest()

    def transform(self):
        return json.dumps(self._program, cls=NumpyEncoder)

    def get_subtree(self, random_state, program=None):
        if program is None:
            program = self._program
        params = program['params']
        # Choice of crossover points follows Koza's (1992) widely used approach
        # of choosing functions 90% of the time and leaves 10% of the time.
        probs = np.array([
            0.9 if params[node] in self._params_sets[node] else 0.1
            for node in params.keys()
        ])
        probs = np.cumsum(probs / probs.sum())
        start = np.searchsorted(probs, random_state.uniform())
        end = start
        while len(list(params.keys())[start:]) > end - start:
            end += 1
        return start, end

    def reproduce(self):
        return copy.deepcopy(self._program)

    def point_mutation(self, random_state):
        program = copy.deepcopy(self._program)
        mutate = np.where(
            random_state.uniform(
                size=len(program['params'])) < self._p_point_replace)[0]
        removed = [
            list(self._program['params'].keys())[node] for node in mutate
        ]
        remain = list(set(self._program['params'].keys()) - set(removed))
        for key in removed:
            program['params'][key] = random_state.choice(
                self._params_sets[key])
        return program, removed, remain

    def crossover(self, donor, random_state):
        start, end = self.get_subtree(random_state)
        removed = list(self._program['params'].keys())[start:end]
        remain = list(set(self._program['params'].keys()) - set(removed))
        program = copy.deepcopy(self._program)
        for key in removed:
            program['params'][key] = donor['params'][key]
        return program, removed, remain

    def raw_fitness(self, features, X, Y, mode, n_splits):
        if self._fitness is None:
            raw_fitness = model_fitness(features=features,
                                        model_name=self._model_name,
                                        X=X,
                                        Y=Y,
                                        mode=mode,
                                        n_splits=n_splits,
                                        params=self._params)
        else:
            raw_fitness = self._fitness(features=features,
                                        model_name=self._model_name,
                                        X=X,
                                        Y=Y,
                                        params=self._params)
        self._raw_fitness = raw_fitness


# -*- encoding:utf-8 -*-
"""封装grid search相关操作模块"""

import logging
import numpy as np
from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.metrics import explained_variance_score, make_scorer

from ultron.ump.core.fixes import signature, six
from ultron.ump.core.fixes import GridSearchCV

__all__ = [
    'grid_search_init_kwargs', 'grid_search_mul_init_kwargs',
    'grid_search_init_n_estimators', 'grid_search_init_n_components',
    'grid_search_init_max_depth', 'grid_search_init_n_neighbors'
]


def _scoring_grid(estimator, scoring):
    """
    只针对有监督学习过滤无监督学习，对scoring未赋予的情况根据
    学习器分类器使用accuracy进行度量，回归器使用可释方差值explained_variance_score，
    使用make_scorer对函数进行score封装

    :param estimator: 学习器对象
    :param scoring: 度量使用的方法，未赋予的情况根据
                    学习器分类器使用accuracy进行度量，回归器使用explained_variance_score进行度量
    :return: scoring
    """
    if not isinstance(estimator, (ClassifierMixin, RegressorMixin)):
        logging.info('only support supervised learning')
        # TODO 无监督学习的scoring度量以及GridSearchCV
        return None

    if scoring is None:
        if isinstance(estimator, ClassifierMixin):
            # 分类器使用accuracy
            return 'accuracy'
        elif isinstance(estimator, RegressorMixin):
            # 回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
            """
                make_scorer中通过greater_is_better对返回值进行正负分配
                eg: sign = 1 if greater_is_better else -1
            """
            return make_scorer(explained_variance_score,
                               greater_is_better=True)
        return None
    return scoring


def grid_search_init_kwargs(estimator,
                            x,
                            y,
                            param_name,
                            param_range,
                            cv=10,
                            n_jobs=-1,
                            scoring=None):
    """
    对GridSearchCV进行封装，对单个目标关键字参数进行grid search最优参数搜寻
        eg：'n_estimators'， 'max_depth'
        eg：param_range=np.arange(100, 500, 50))对最优参数进行寻找

    eg:
        from ultronpy import UltronML, ml
        ttn_ultron = UltronML.create_test_more_fiter()
        ttn_ultron.estimator.random_forest_classifier()
        ml.grid_search_init_kwargs(ttn_ultron.estimator.clf, ttn_ultron.x, ttn_ultron.y,
                           param_name='n_estimators', param_range=np.arange(100, 500, 50))

        可找到n_estimators参数最优为：(0.81930415263748602, {'n_estimators': 300})

    :param estimator: 学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param param_name: 做为grid的目标关键字参数，eg：'n_estimators'， 'max_depth'
    :param param_range: 做为grid的目标关键字参数的grid序列，eg：param_range=np.arange(100, 500, 50))
    :param cv: int，GridSearchCV切割训练集测试集参数，默认10
    :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
    :param scoring: 测试集的度量方法，默认为None, None的情况下分类器使用accuracy进行度量，
                    回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
    :param show: 是否进行可视化
    :return: eg：(0.81930415263748602, {'n_estimators': 300})
    """
    if not isinstance(param_name, six.string_types):
        # param_name参数需要是字符串类型
        logging.info('param_name is str, not {}, eg: \'n_estimators\''.format(
            param_name))
        return None, None

    # 根据分类回归得到非None的score
    scoring = _scoring_grid(estimator, scoring)
    if scoring is None:
        # 如果_scoring_grid返回的结果仍然是None, 说明无监督学习，暂时不支持
        return None, None

    # 获取学习器的init函数，使用getattr
    init = getattr(estimator.__class__.__init__, 'deprecated_original',
                   estimator.__class__.__init__)

    # 获取函数签名
    init_signature = signature(init)
    """
        eg：init_signature
            ['self', 'base_estimator', 'n_estimators', 'max_samples', 'max_features', 'bootstrap',
            'bootstrap_features', 'oob_score', 'warm_start', 'n_jobs', 'random_state', 'verbose']
    """
    if param_name not in init_signature.parameters.keys():
        # 如果需要grid的参数param_name不在init函数签名中，打log，返回
        logging.info(
            'check init signature {} not in **kwargs\ninit_signature:{}'.
            format(param_name, init_signature.parameters.keys()))
        return None, None

    param_grid = {param_name: param_range}
    grid = GridSearchCV(estimator,
                        param_grid,
                        cv=cv,
                        scoring=scoring,
                        n_jobs=n_jobs)
    print('start grid search please wait...')
    grid.fit(x, y)
    return grid.best_score_, grid.best_params_


def grid_search_mul_init_kwargs(estimator,
                                x,
                                y,
                                param_grid,
                                cv=10,
                                n_jobs=-1,
                                scoring=None):
    """
    对GridSearchCV进行封装，对多个目标关键字参数进行grid search最优参数搜寻

    eg:
        from ultronpy import UltronML, ml
        ttn_ultron = UltronML.create_test_more_fiter()
        ttn_ultron.estimator.random_forest_classifier()

        param_grid = {'max_depth': np.arange(2, 5), 'n_estimators': np.arange(100, 300, 50)}
        ml.grid_search_mul_init_kwargs(ttn_ultron.estimator.clf, ttn_ultron.x, ttn_ultron.y, param_grid=param_grid)

        out: (0.81593714927048255, {'max_depth': 4, 'n_estimators': 250})

    :param estimator: 学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param param_grid: eg：param_grid = {'max_depth': np.arange(2, 5), 'n_estimators': np.arange(100, 300, 50)}
    :param cv: int，GridSearchCV切割训练集测试集参数，默认10
    :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
    :param scoring: 测试集的度量方法，默认为None, None的情况下分类器使用accuracy进行度量，
                    回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
    :return: eg: (0.81593714927048255, {'max_depth': 4, 'n_estimators': 250})
    """
    if not isinstance(param_grid, dict):
        # param_grid参数是dict对象
        logging.info('param_grid is dict object, not {}}'.format(param_grid))
        return None, None

    # 根据分类回归得到非None的score
    scoring = _scoring_grid(estimator, scoring)
    if scoring is None:
        # 如果_scoring_grid返回的结果仍然是None, 说明无监督学习，暂时不支持
        return None, None

    for param_name in param_grid.keys():
        # 迭代每一个key，即每一个关键字参数，看init方法中的签名是否存在该关键字

        # 获取学习器的init函数，使用getattr
        init = getattr(estimator.__class__.__init__, 'deprecated_original',
                       estimator.__class__.__init__)

        # 获取函数签名
        init_signature = signature(init)
        if param_name not in init_signature.parameters.keys():
            # 如果需要grid的参数param_name不在init函数签名中，打log，返回
            logging.info(
                'check init signature {} not in **kwargs\ninit_signature:{}'.
                format(param_name, init_signature.parameters.keys()))
            return None, None

    grid = GridSearchCV(estimator,
                        param_grid,
                        cv=cv,
                        scoring=scoring,
                        n_jobs=n_jobs)
    print('start grid search please wait...')
    grid.fit(x, y)
    return grid.best_score_, grid.best_params_


def grid_search_init_n_estimators(estimator,
                                  x,
                                  y,
                                  n_estimators_range=None,
                                  cv=10,
                                  n_jobs=-1,
                                  scoring=None):
    """
    封装grid search特定的'n_estimators'关键字参数最优搜索，
    为UltronMLCreater中_estimators_prarms_best提供callback函数，

    具体阅读
            UltronMLCreater._estimators_prarms_best()
            + UltronMLCreater.random_forest_classifier_best()

    eg:
        from ultronpy import UltronML, ml
        ttn_ultron = UltronML.create_test_more_fiter()
        ttn_ultron.estimator.random_forest_classifier()
        ml.grid_search_init_n_estimators(ttn_ultron.estimator.clf, ttn_ultron.x, ttn_ultron.y)

    :param estimator: 学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param n_estimators_range: 默认None, None则会使用 n_estimators_range = np.arange(50, 500, 10)
    :param cv: int，GridSearchCV切割训练集测试集参数，默认10
    :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
    :param scoring: 测试集的度量方法，默认为None, None的情况下分类器使用accuracy进行度量，
                    回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
    :param show: 是否进行可视化
    :return: eg: (0.82154882154882158, {'n_estimators': 310})
    """

    if n_estimators_range is None:
        n_estimators_range = np.arange(50, 500, 10)

    return grid_search_init_kwargs(estimator,
                                   x,
                                   y,
                                   'n_estimators',
                                   n_estimators_range,
                                   cv=cv,
                                   n_jobs=n_jobs,
                                   scoring=scoring)


def grid_search_init_max_depth(estimator,
                               x,
                               y,
                               max_depth_range=None,
                               cv=10,
                               n_jobs=-1,
                               scoring=None):
    """
    封装grid search特定的'n_components'关键字参数最优搜索，
    为UltronMLCreater中_estimators_prarms_best提供callback函数

    具体阅读
            UltronMLCreater._estimators_prarms_best()
            + UltronMLCreater.decision_tree_classifier_best()

    :param estimator: 学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param max_depth_range: 默认None, None则会使用:
            max_depth_range = np.arange(2, np.maximum(10, int(x.shape[1]) - 1), 1)

    :param cv: int，GridSearchCV切割训练集测试集参数，默认10
    :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
    :param scoring: 测试集的度量方法，默认为None, None的情况下分类器使用accuracy进行度量，
                    回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
    :param show: 是否进行可视化
    :return: eg: (0.82154882154882158, {'max_depth': 3})
    """

    if max_depth_range is None:
        max_depth_range = np.arange(2, np.maximum(10, int(x.shape[1]) - 1), 1)

    return grid_search_init_kwargs(estimator,
                                   x,
                                   y,
                                   'max_depth',
                                   max_depth_range,
                                   cv=cv,
                                   n_jobs=n_jobs,
                                   scoring=scoring)


def grid_search_init_n_neighbors(estimator,
                                 x,
                                 y,
                                 n_neighbors_range=None,
                                 cv=10,
                                 n_jobs=-1,
                                 scoring=None):
    """
    封装grid search特定的'n_components'关键字参数最优搜索，
    为UltronMLCreater中_estimators_prarms_best提供callback函数

    具体阅读
            UltronMLCreater._estimators_prarms_best()
            + UltronMLCreater.knn_classifier_best()

    :param estimator: 学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param n_neighbors_range: 默认None, None则会使用:
            n_estimators_range = np.arange(2, np.maximum(10, int(x.shape[1]) - 1), 1)

    :param cv: int，GridSearchCV切割训练集测试集参数，默认10
    :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
    :param scoring: 测试集的度量方法，默认为None, None的情况下分类器使用accuracy进行度量，
                    回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
    :param show: 是否进行可视化
    :return: eg: (0.82154882154882158, {'n_components': 10})
    """

    if n_neighbors_range is None:
        # 邻居投票者控制在1-np.minimum(26, 总数的1/3）
        n_neighbors_range = np.arange(1, np.minimum(26, int(x.shape[0] / 3)),
                                      1)

    return grid_search_init_kwargs(estimator,
                                   x,
                                   y,
                                   'n_neighbors',
                                   n_neighbors_range,
                                   cv=cv,
                                   n_jobs=n_jobs,
                                   scoring=scoring)


def grid_search_init_n_components(estimator,
                                  x,
                                  y,
                                  n_components_range=None,
                                  cv=10,
                                  n_jobs=-1,
                                  scoring=None):
    """
    封装grid search特定的'n_components'关键字参数最优搜索，
    为UltronMLCreater中_estimators_prarms_best提供callback函数，
    具体阅读UltronMLCreater._estimators_prarms_best()

    :param estimator: 学习器对象
    :param x: 训练集x矩阵，numpy矩阵
    :param y: 训练集y序列，numpy序列
    :param n_components_range: 默认None, None则会使用:
            n_estimators_range = np.arange(2, np.maximum(10, int(x.shape[1]) - 1), 1)

    :param cv: int，GridSearchCV切割训练集测试集参数，默认10
    :param n_jobs: 并行执行的进程任务数量，默认-1, 开启与cpu相同数量的进程数
    :param scoring: 测试集的度量方法，默认为None, None的情况下分类器使用accuracy进行度量，回归器使用
                    回归器使用可释方差值explained_variance_score，使用make_scorer对函数进行score封装
    :param show: 是否进行可视化
    :return: eg: (0.82154882154882158, {'n_components': 10})
    """
    if n_components_range is None:
        n_components_range = np.arange(2, np.maximum(10,
                                                     int(x.shape[1]) - 1), 1)

    return grid_search_init_kwargs(estimator,
                                   x,
                                   y,
                                   'n_components',
                                   n_components_range,
                                   cv=cv,
                                   n_jobs=n_jobs,
                                   scoring=scoring)



# -*- coding: utf-8 -*-


import copy
import bisect
import six
import numpy as np
cimport numpy as np
cimport cython
from libc.math cimport isnan
from libc.math cimport log
from libc.math cimport sqrt
from copy import deepcopy
from ultron.sentry.Math.Accumulators.IAccumulators cimport Accumulator
from ultron.sentry.Math.Accumulators.IAccumulators cimport Latest
from ultron.sentry.Math.Accumulators.IAccumulators cimport build_holder
from ultron.sentry.Math.Accumulators.StatelessAccumulators cimport PositivePart
from ultron.sentry.Math.Accumulators.StatelessAccumulators cimport NegativePart
from ultron.sentry.Math.Accumulators.StatelessAccumulators cimport XAverage
from ultron.sentry.Math.Accumulators.IAccumulators cimport Pow
from ultron.sentry.Utilities.Asserts cimport pyFinAssert
from ultron.sentry.Utilities.Asserts cimport isClose
from ultron.sentry.Math.udfs cimport consecutive_int_sum
from ultron.sentry.Math.Accumulators.impl cimport Deque
from ultron.sentry.Math.MathConstants cimport NAN


cdef class StatefulValueHolder(Accumulator):

    def __init__(self, window):
        super(StatefulValueHolder, self).__init__()
        if not isinstance(window, int):
            raise ValueError("window parameter should be a positive int however {0} received"
                             .format(window))
        pyFinAssert(window > 0, ValueError, "window length should be greater than 0")
        self._deque = Deque(window)
        self._isFull = False

    cpdef size_t size(self):
        return self._deque.size()

    cpdef bint isFull(self):
        return self._isFull


cdef class Shift(StatefulValueHolder):

    def __init__(self, window, x):
        super(Shift, self).__init__(window)
        pyFinAssert(window >= 1, ValueError, "shift value should not be less than 1")
        self._x = build_holder(x)
        self._window = self._x.window + window
        self._dependency = deepcopy(self._x.dependency)
        self._popout = NAN

    cpdef push(self, dict data):
        self._x.push(data)
        self._popout = self._deque.dump(self._x.result())
        self._isFull = self._isFull or self._deque.isFull()

    cpdef double result(self):
        return self._popout

    cpdef int lag(self):
        return self._window - self._x.window

    def __str__(self):
        return '\mathrm{{Shift}}({0}, {1})'.format(str(self._x), self._window - self._x.window)


cdef class Delta(StatefulValueHolder):

    def __init__(self, window, x):
        super(Delta, self).__init__(window)
        pyFinAssert(window >= 1, ValueError, "Delta window value should not be less than 1")
        self._x = build_holder(x)
        self._window = self._x.window + window
        self._dependency = deepcopy(self._x.dependency)
        self._popout = NAN
        self._current = NAN

    cpdef push(self, dict data):
        self._x.push(data)
        self._current = self._x.result()
        self._popout = self._deque.dump(self._current)
        self._isFull = self._isFull or self._deque.isFull()

    cpdef double result(self):
        if self._isFull:
            return self._current - self._popout
        else:
            return NAN

    cpdef int lag(self):
        return self._window - self._x.window

    def __str__(self):
        return '\mathrm{{Delta}}({0}, {1})'.format(str(self._x), self._window - self._x.window)


cdef class SingleValuedValueHolder(StatefulValueHolder):
    def __init__(self, window, x):
        super(SingleValuedValueHolder, self).__init__(window)
        self._x = build_holder(x)
        self._window = self._x.window + window
        self._dependency = deepcopy(self._x.dependency)


cdef class SortedValueHolder(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(SortedValueHolder, self).__init__(window, x)
        self._sortedArray = []
        self._cur_pos = NAN

    cpdef push(self, dict data):
        cdef double popout
        cdef int delPos

        self._x.push(data)
        cdef double value = self._x.result()
        if isnan(value):
            return NAN
        if self._deque.isFull():
            popout = self._deque.dump(value)
            delPos = bisect.bisect_left(self._sortedArray, popout)
            del self._sortedArray[delPos]
            self._cur_pos = bisect.bisect_left(self._sortedArray, value)
            self._sortedArray.insert(int(self._cur_pos), value)
        else:
            self._deque.dump(value)
            self._cur_pos = bisect.bisect_left(self._sortedArray, value)
            self._sortedArray.insert(int(self._cur_pos), value)
        self._isFull = self._isFull or self._deque.isFull()


cdef class MovingMax(SortedValueHolder):
    def __init__(self, window, x):
        super(MovingMax, self).__init__(window, x)

    cpdef double result(self):
        if self._sortedArray:
            return self._sortedArray[-1]
        else:
            return NAN

    def __str__(self):
        return "\\mathrm{{MMax}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingArgMax(SortedValueHolder):

    def __init__(self, window, x):
        super(MovingArgMax, self).__init__(window, x)

    cpdef double result(self):
        cdef double currMax
        cdef double idx
        if self._sortedArray:
            currMax = self._sortedArray[-1]
            idx = self._deque.idx(currMax)
            return self.size() - idx - 1
        else:
            return NAN

    def __str__(self):
        return "\\mathrm{{MArgMax}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingMin(SortedValueHolder):
    def __init__(self, window, x):
        super(MovingMin, self).__init__(window, x)

    cpdef double result(self):
        if self._sortedArray:
            return self._sortedArray[0]
        else:
            return NAN

    def __str__(self):
        return "\\mathrm{{MMin}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingArgMin(SortedValueHolder):

    def __init__(self, window, x):
        super(MovingArgMin, self).__init__(window, x)

    cpdef double result(self):
        cdef double currMax
        cdef double idx
        if self._sortedArray:
            currMin = self._sortedArray[0]
            idx = self._deque.idx(currMin)
            return self.size() - idx - 1
        else:
            return NAN

    def __str__(self):
        return "\\mathrm{{MArgMin}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingRank(SortedValueHolder):
    def __init__(self, window, x):
        super(MovingRank, self).__init__(window, x)

    cpdef double result(self):
        return self._cur_pos

    def __str__(self):
        return "\\mathrm{{MRank}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingQuantile(SortedValueHolder):
    def __init__(self, window, x):
        super(MovingQuantile, self).__init__(window, x)

    @cython.cdivision(True)
    cpdef double result(self):
        cdef size_t n = len(self._sortedArray)
        if n > 1:
            return self._cur_pos / (n - 1)
        else:
            return NAN

    def __str__(self):
        return "\\mathrm{{MQuantile}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingAllTrue(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingAllTrue, self).__init__(window, x)
        self._countedTrue = 0

    cpdef push(self, dict data):
        cdef int addedTrue
        cdef double popout

        self._x.push(data)
        cdef double value = self._x.result()
        if isnan(value):
            return NAN
        addedTrue = 0

        if value:
            addedTrue += 1
        popout = self._deque.dump(value, False)
        if popout:
            addedTrue -= 1

        self._countedTrue += addedTrue
        self._isFull = self._isFull or self._deque.isFull()

    cpdef double result(self):
        return self._countedTrue == self.size()

    def __str__(self):
        return "\\mathrm{{MAllTrue}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingAnyTrue(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingAnyTrue, self).__init__(window, x)
        self._countedTrue = 0

    cpdef push(self, dict data):
        cdef int addedTrue
        cdef double popout

        self._x.push(data)
        cdef double value = self._x.result()
        if isnan(value):
            return NAN
        addedTrue = 0

        if value:
            addedTrue += 1
        popout = self._deque.dump(value, False)
        if popout:
            addedTrue -= 1

        self._countedTrue += addedTrue
        self._isFull = self._isFull or self._deque.isFull()

    cpdef double result(self):
        return self._countedTrue != 0

    def __str__(self):
        return "\\mathrm{{MAnyTrue}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingSum(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingSum, self).__init__(window, x)
        self._runningSum = 0.0

    cpdef push(self, dict data):
        cdef double popout
        self._x.push(data)
        cdef double value = self._x.result()
        if isnan(value):
            return NAN
        popout = self._deque.dump(value, 0.)
        self._runningSum = self._runningSum - popout + value

        self._isFull = self._isFull or self._deque.isFull()

    cpdef double result(self):
        return self._runningSum

    def __str__(self):
        return "\\mathrm{{MSum}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingAverage(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingAverage, self).__init__(window, x)
        self._runningSum = 0.0

    cpdef push(self, dict data):
        cdef double popout

        self._x.push(data)
        cdef double value = self._x.result()

        if isnan(value):
            return NAN
        popout = self._deque.dump(value, 0.)
        self._runningSum += value - popout
        self._isFull = self._isFull or self._deque.isFull()

    @cython.cdivision(True)
    cpdef double result(self):
        cdef size_t size = self.size()
        if size:
            return self._runningSum / size
        else:
            return NAN

    def __str__(self):
        return "\\mathrm{{MA}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingDecay(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingDecay, self).__init__(window, x)
        self._runningSum = 0.0
        self._runningWeightedSum = 0.0
        self._newestValue = 0.

    cpdef push(self, dict data):
        cdef double popout
        cdef double k = self._window

        self._x.push(data)
        cdef double value = self._x.result()

        if isnan(value):
            return NAN

        popout = self._deque.dump(value, 0.)
        self._runningWeightedSum = self._runningWeightedSum - self._runningSum + k * value
        self._runningSum += value - popout
        self._newestValue = value
        self._isFull = self._isFull or self._deque.isFull()

    @cython.cdivision(True)
    cpdef double result(self):
        cdef double k = self._window
        cdef size_t s = self.size()
        cdef double c = consecutive_int_sum(k - s + 1, k)
        if s:
            return self._runningWeightedSum / c
        else:
            return NAN

    def __str__(self):
        return "\\mathrm{{MADecay}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingPositiveAverage(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingPositiveAverage, self).__init__(window, x)
        self._runningPositiveSum = 0.0
        self._runningPositiveCount = 0

    cpdef push(self, dict data):
        cdef double popout

        self._x.push(data)
        cdef double value = self._x.result()

        if isnan(value):
            return NAN
        popout = self._deque.dump(value, -1.)
        if value > 0.0:
            self._runningPositiveCount += 1
            self._runningPositiveSum += value

        if popout > 0.0:
            self._runningPositiveCount -= 1
            self._runningPositiveSum -= popout
        self._isFull = self._isFull or self._deque.isFull()

    @cython.cdivision(True)
    cpdef double result(self):
        if self._runningPositiveCount == 0:
            return 0.0
        else:
            return self._runningPositiveSum / self._runningPositiveCount

    def __str__(self):
        return "\\mathrm{{mposavg}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingPositiveDifferenceAverage(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingPositiveDifferenceAverage, self).__init__(window, x)
        cdef Accumulator runningPositive = PositivePart(build_holder(x) - Shift(1, build_holder(x)))
        self._runningAverage = MovingAverage(window, x=runningPositive)

    cpdef push(self, dict data):
        self._runningAverage.push(data)
        self._isFull = self._isFull or self._runningAverage.isFull()

    cpdef double result(self):
        return self._runningAverage.result()

    def __str__(self):
        return "\\mathrm{{mposdiffavg}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingNegativeDifferenceAverage(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingNegativeDifferenceAverage, self).__init__(window, x)
        cdef Accumulator runningNegative = NegativePart(build_holder(x) - Shift(1, build_holder(x)))
        self._runningAverage = MovingAverage(window, x=runningNegative)

    cpdef push(self, dict data):
        self._runningAverage.push(data)
        self._isFull = self._isFull or self._runningAverage.isFull()

    cpdef double result(self):
        return self._runningAverage.result()

    def __str__(self):
        return "\\mathrm{{mnegdiffavg}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingRSI(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingRSI, self).__init__(window, x)
        self._posDiffAvg = MovingPositiveDifferenceAverage(window, x)
        self._negDiffAvg = MovingNegativeDifferenceAverage(window, x)

    cpdef push(self, dict data):
        self._posDiffAvg.push(data)
        self._negDiffAvg.push(data)
        self._isFull = self._isFull or (self._posDiffAvg.isFull() and self._negDiffAvg.isFull())

    @cython.cdivision(True)
    cpdef double result(self):
        cdef double nominator = self._posDiffAvg.result()
        cdef double denominator = nominator - self._negDiffAvg.result()
        if denominator:
            return 100. * nominator / denominator
        else:
            return 50.

    def __str__(self):
        return "\\mathrm{{mrsi}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingNegativeAverage(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingNegativeAverage, self).__init__(window, x)
        self._runningNegativeSum = 0.0
        self._runningNegativeCount = 0

    cpdef push(self, dict data):
        cdef double popout

        self._x.push(data)
        cdef double value = self._x.result()

        if isnan(value):
            return NAN
        popout = self._deque.dump(value, 1.)
        if value < 0.0:
            self._runningNegativeCount += 1
            self._runningNegativeSum += value

        if popout < 0.0:
            self._runningNegativeCount -= 1
            self._runningNegativeSum -= popout
        self._isFull = self._isFull or self._deque.isFull()

    @cython.cdivision(True)
    cpdef double result(self):
        if self._runningNegativeCount:
            return self._runningNegativeSum / self._runningNegativeCount
        else:
            return 0.

    def __str__(self):
        return "\\mathrm{{mnegavg}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingVariance(SingleValuedValueHolder):

    def __init__(self, window, x, isPopulation=False):
        super(MovingVariance, self).__init__(window, x)
        self._runningSum = 0.0
        self._runningSumSquare = 0.0
        self._isPop = isPopulation
        if not self._isPop:
            pyFinAssert(window >= 2, ValueError, "sampling variance can't be calculated with window size < 2")

    cpdef push(self, dict data):
        cdef double popout
        self._x.push(data)
        cdef double value = self._x.result()
        if isnan(value):
            return NAN
        popout = self._deque.dump(value, 0.)

        self._runningSum += value - popout
        self._runningSumSquare += value * value - popout * popout
        self._isFull = self._isFull or self._deque.isFull()

    @cython.cdivision(True)
    cpdef double result(self):
        cdef size_t length = self._deque.size()
        cdef double tmp

        if length == 0:
            return NAN

        tmp = self._runningSumSquare - self._runningSum * self._runningSum / length

        if self._isPop:
            return tmp / length
        else:
            if length >= 2:
                return tmp / (length - 1)
            else:
                return NAN

    def __str__(self):
        return "\\mathrm{{mvar}}({0}, {1}, {2})".format(self._window, str(self._x), self._isPop)


cdef class MovingStandardDeviation(SingleValuedValueHolder):

    def __init__(self, window, x, isPopulation=False):
        super(MovingStandardDeviation, self).__init__(window, x)
        self._runningSum = 0.0
        self._runningSumSquare = 0.0
        self._isPop = isPopulation
        if not self._isPop:
            pyFinAssert(window >= 2, ValueError, "sampling standard deviation can't be calculated with window size < 2")

    cpdef push(self, dict data):
        cdef double popout
        self._x.push(data)
        cdef double value = self._x.result()

        if isnan(value):
            return NAN
        popout = self._deque.dump(value, 0.)

        self._runningSum += value - popout
        self._runningSumSquare += value * value - popout * popout
        self._isFull = self._isFull or self._deque.isFull()

    @cython.cdivision(True)
    cpdef double result(self):
        cdef size_t length = self._deque.size()
        cdef double tmp

        if length == 0:
            return NAN

        tmp = self._runningSumSquare - self._runningSum * self._runningSum / length

        if self._isPop:
            return sqrt(tmp / length)
        else:
            if length >= 2:
                return sqrt(tmp / (length - 1))
            else:
                return NAN

    def __str__(self):
        return "\\mathrm{{mstd}}({0}, {1}, {2})".format(self._window, str(self._x), self._isPop)


cdef class MovingNegativeVariance(SingleValuedValueHolder):

    def __init__(self, window, x, isPopulation=0):
        super(MovingNegativeVariance, self).__init__(window, x)
        self._runningNegativeSum = 0.0
        self._runningNegativeSumSquare = 0.0
        self._runningNegativeCount = 0
        self._isPop = isPopulation
        if not self._isPop:
            pyFinAssert(window >= 2, ValueError, "sampling standard deviation can't be calculated with window size < 2")

    cpdef push(self, dict data):
        cdef double popout
        self._x.push(data)
        cdef double value = self._x.result()

        if isnan(value):
            return NAN
        popout = self._deque.dump(value, 1.)

        if value < 0:
            self._runningNegativeSum += value
            self._runningNegativeSumSquare += value * value
            self._runningNegativeCount += 1
        if popout < 0:
            self._runningNegativeSum -= popout
            self._runningNegativeSumSquare -= popout * popout
            self._runningNegativeCount -= 1
        self._isFull = self._isFull or self._deque.isFull()

    @cython.cdivision(True)
    cpdef double result(self):

        cdef int length
        cdef double tmp

        if self._isPop:
            if self._runningNegativeCount >= 1:
                length = self._runningNegativeCount
                tmp = self._runningNegativeSumSquare - self._runningNegativeSum * self._runningNegativeSum / length
                return tmp / length
            else:
                return NAN
        else:
            if self._runningNegativeCount >= 2:
                length = self._runningNegativeCount
                tmp = self._runningNegativeSumSquare - self._runningNegativeSum * self._runningNegativeSum / length
                return tmp / (length - 1)
            else:
                return NAN

    def __str__(self):
        return "\\mathrm{{mnegvar}}({0}, {1}, {2})".format(self._window, str(self._x), self._isPop)


cdef class MovingCountedPositive(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingCountedPositive, self).__init__(window, x)
        self._counts = 0

    cpdef push(self, dict data):
        cdef double popout
        self._x.push(data)
        cdef double value = self._x.result()

        if isnan(value):
            return NAN
        popout = self._deque.dump(value, -1.)

        if value > 0:
            self._counts += 1
        if popout > 0:
            self._counts -= 1
        self._isFull = self._isFull or self._deque.isFull()

    cpdef double result(self):
        return self._counts

    def __str__(self):
        return "\\mathrm{{mposcount}}({0}, {1})".format(self._window, str(self._x))


cdef class MovingCountedNegative(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingCountedNegative, self).__init__(window, x)
        self._counts = 0

    cpdef push(self, dict data):
        cdef double popout
        self._x.push(data)
        cdef double value = self._x.result()

        if isnan(value):
            return NAN
        popout = self._deque.dump(value, 1)

        if value < 0:
            self._counts += 1
        if popout < 0:
            self._counts -= 1
        self._isFull = self._isFull or self._deque.isFull()

    cpdef double result(self):
        return self._counts

    def __str__(self):
        return "\\mathrm{{mnegcount}}({0}, {1})".format(self._window, str(self._x))


# Calculator for one pair of series
cdef class MovingCorrelation(StatefulValueHolder):

    def __init__(self, window, x, y):
        super(MovingCorrelation, self).__init__(window)
        self._runningSumLeft = 0.0
        self._runningSumRight = 0.0
        self._runningSumSquareLeft = 0.0
        self._runningSumSquareRight = 0.0
        self._runningSumCrossSquare = 0.0
        self._x = build_holder(x)
        self._y = build_holder(y)
        self._window = max(self._x.window + window, self._y.window + window)
        self._dependency = list(set(self._x.dependency + self._y.dependency))
        self._deque_y = Deque(window)

    cpdef push(self, dict data):
        cdef double headLeft
        cdef double headRight

        self._x.push(data)
        cdef double x = self._x.result()

        self._y.push(data)
        cdef double y = self._y.result()

        if isnan(x) or isnan(y):
            return NAN

        headLeft = self._deque.dump(x, 0.)
        headRight = self._deque_y.dump(y, 0.)

        # updating cached values
        self._runningSumLeft = self._runningSumLeft - headLeft + x
        self._runningSumRight = self._runningSumRight - headRight + y
        self._runningSumSquareLeft = self._runningSumSquareLeft - headLeft * headLeft + x * x
        self._runningSumSquareRight = self._runningSumSquareRight - headRight * headRight + y * y
        self._runningSumCrossSquare = self._runningSumCrossSquare - headLeft * headRight + x * y
        self._isFull = self._isFull or self._deque.isFull()

    cpdef double result(self):
        cdef size_t n = self.size()
        cdef double nominator
        cdef double denominator
        if n >= 2:
            nominator = n * self._runningSumCrossSquare - self._runningSumLeft * self._runningSumRight
            denominator = (n * self._runningSumSquareLeft - self._runningSumLeft * self._runningSumLeft) \
                          * (n * self._runningSumSquareRight - self._runningSumRight * self._runningSumRight)
            if not isClose(denominator, 0.):
                denominator = sqrt(denominator)
                return nominator / denominator
            else:
                return 0.0
        else:
            return NAN

    def __str__(self):
        return "\\mathrm{{mcorr}}({0}, {1}, {2})".format(self._window, str(self._x), str(self._y))


cdef class MovingProduct(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingProduct, self).__init__(window, x)
        self._runningProduct = 1.0

    cpdef push(self, dict data):
        cdef double popout
        self._x.push(data)
        cdef double value = self._x.result()
        if isnan(value):
            return NAN
        popout = self._deque.dump(value)
        if not isnan(popout):
            self._runningProduct *= value / popout
        else:
            self._runningProduct *= value
        self._isFull = self._isFull or self._deque.isFull()

    cpdef double result(self):
        return self._runningProduct

    def __str__(self):
        return "\\mathrm{{mprod}}({0}, {1})".format(self._window, str(self._x))


cdef class MACD(Accumulator):

    def __init__(self, short_win, long_win, x, method=XAverage):
        super(MACD, self).__init__()
        self._short_average = method(window=short_win, x=x)
        self._long_average = method(window=long_win, x=x)
        self._isFull = False
        self._window = max(self._short_average.window, self._long_average.window)
        self._dependency = self._short_average._dependency

    cpdef push(self, dict data):
        self._short_average.push(data)
        self._long_average.push(data)
        self._isFull = self._isFull or (self._short_average.isFull() and self._long_average.isFull())

    cpdef double result(self):
        return self._short_average.result() - self._long_average.result()


'''
performancer
'''

cdef class MovingLogReturn(SingleValuedValueHolder):

    def __init__(self, window, x):
        super(MovingLogReturn, self).__init__(window, x)
        self._runningReturn = NAN

    @cython.cdivision(True)
    cpdef push(self, dict data):
        cdef double popout
        self._x.push(data)
        cdef double value = self._x.result()
        if isnan(value):
            return NAN
        popout = self._deque.dump(value)
        if not isnan(popout):
            self._runningReturn = log(value / popout)
        self._isFull = self._isFull or self._deque.isFull()

    cpdef double result(self):
        return self._runningReturn


cdef class MovingSharp(StatefulValueHolder):

    def __init__(self, window, x, y):
        super(MovingSharp, self).__init__(window)
        self._mean = MovingAverage(window, x='x')
        self._var = MovingVariance(window, x='x', isPopulation=False)
        self._x = build_holder(x)
        self._y = build_holder(y)
        self._window = max(self._x.window + window, self._y.window + window)
        self._dependency = list(set(self._x.dependency + self._y.dependency))
        self._deque_y = Deque(window)

    cpdef push(self, dict data):

        cdef tuple value
        cdef double ret
        cdef double benchmark
        cdef dict new_data

        self._x.push(data)
        ret = self._x.result()

        self._y.push(data)
        benchmark = self._y.result()

        if isnan(ret) or isnan(benchmark):
            return NAN

        new_data = {'x': ret - benchmark}
        self._mean.push(new_data)
        self._var.push(new_data)
        self._isFull = self._isFull or (self._mean.isFull() and self._var.isFull())

    @cython.cdivision(True)
    cpdef double result(self):
        cdef double tmp = self._var.result()
        if tmp != 0.:
            return self._mean.result() / sqrt(tmp)
        else:
            return NAN


cdef class MovingSortino(StatefulValueHolder):

    def __init__(self, window, x, y):
        super(MovingSortino, self).__init__(window)
        self._mean = MovingAverage(window, x='x')
        self._negativeVar = MovingNegativeVariance(window, x='x')
        self._x = build_holder(x)
        self._y = build_holder(y)
        self._window = max(self._x.window + window, self._y.window + window)
        self._dependency = list(set(self._x.dependency + self._y.dependency))
        self._deque_y = Deque(window)

    cpdef push(self, dict data):
        cdef tuple value
        cdef double ret
        cdef double benchmark
        cdef dict new_data

        self._x.push(data)
        ret = self._x.result()

        self._y.push(data)
        benchmark = self._y.result()

        if isnan(ret) or isnan(benchmark):
            return NAN

        new_data = {'x': ret - benchmark}
        self._mean.push(new_data)
        self._negativeVar.push(new_data)
        self._isFull = self._isFull or (self._negativeVar.isFull() and self._mean.isFull())

    @cython.cdivision(True)
    cpdef double result(self):
        cdef double tmp = self._negativeVar.result()
        if tmp != 0.:
            return self._mean.result() /sqrt(self._negativeVar.result())
        else:
            return NAN


cdef class MovingDrawdown(StatefulValueHolder):

    def __init__(self, window, x):
        super(MovingDrawdown, self).__init__(window)
        self._maxer = MovingMax(window + 1, x='x')
        self._maxer.push(dict(x=0.0))
        self._x = build_holder(x)
        self._runningCum = 0.0
        self._currentMax = NAN

    cpdef push(self, dict data):
        self._x.push(data)
        cdef double ret = self._x.result()

        if isnan(ret):
            return NAN

        self._runningCum += ret
        self._maxer.push(dict(x=self._runningCum))
        self._currentMax = self._maxer.result()
        self._isFull = self._isFull or self._maxer.isFull()

    cpdef double result(self):
        return self._runningCum - self._currentMax


cdef class MovingMaxDrawdown(StatefulValueHolder):

    def __init__(self, window, x):
        super(MovingMaxDrawdown, self).__init__(window)
        self._drawdownCalculator = MovingDrawdown(window, 'x')
        self._minimer = MovingMin(window, 'x')
        self._x = build_holder(x)

    cpdef push(self, dict data):
        self._x.push(data)
        cdef double ret = self._x.result()

        if isnan(ret):
            return NAN

        self._drawdownCalculator.push(dict(x=ret))
        cdef double draw_down = self._drawdownCalculator.result()
        self._minimer.push(dict(x=draw_down))
        self._deque.dump(draw_down)
        self._isFull = self._isFull or self._deque.isFull()

    cpdef double result(self):
        return self._minimer.result()


cdef class MovingResidue(StatefulValueHolder):

    def __init__(self, window, x, y):
        super(MovingResidue, self).__init__(window)
        self._cross = 0.
        self._xsquare = 0.
        self._lasty = NAN
        self._lastx = NAN
        self._x = build_holder(x)
        self._y = build_holder(y)
        self._window = max(self._x.window + window, self._y.window + window)
        self._dependency = list(set(self._x.dependency + self._y.dependency))
        self._deque_y = Deque(window)

    @cython.boundscheck(False)
    @cython.wraparound(False)
    cpdef push(self, dict data):
        cdef double x
        cdef double y
        cdef double head_y
        cdef double head_x

        self._x.push(data)
        x = self._x.result()
        self._y.push(data)
        y = self._y.result()
        if isnan(x) or isnan(y):
            return NAN

        head_x = self._deque.dump(x, 0.)
        head_y = self._deque_y.dump(y, 0.)

        self._cross += x * y - head_x * head_y
        self._xsquare += x * x - head_x * head_x

        self._lastx = x
        self._lasty = y

    cpdef bint isFull(self):
        return self._deque.isFull()

    @cython.cdivision(True)
    cpdef double result(self):
        cdef double coef = self._cross / self._xsquare
        return self._lasty - coef * self._lastx

    def __str__(self):
        return "\\mathrm{{Res}}({0}, {1}, {2})".format(self._window, str(self._x), str(self._y))




# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import copy
import ultron.factor.empyrical as empyrical
from ultron.strategy.executor.naive import NaiveExecutor
from ultron.utilities.logger import kd_logger
from ultron.tradingday import *
from ultron.factor.analysis.factor_analysis import er_portfolio_analysis
from ultron.factor.othgnz.othgnz_engine import OthgnzEngine
from ultron.factor.combine.combine_engine import CombineEngine
from ultron.factor.data.processing import factor_processing
from ultron.factor.data.standardize import standardize
from ultron.factor.data.winsorize import winsorize_normal
from ultron.optimize.model.linearmodel import ConstLinearModel
from ultron.optimize.constraints import LinearConstraints, \
    create_box_bounds, BoundaryType


def create_params(**kwargs):
    turn_over_target = 1.0 if 'turn_over_target' not in kwargs else kwargs[
        'turn_over_target']
    target_vol = 0.1 if 'target_vol' not in kwargs else kwargs['target_vol']
    lbound = 0. if 'lbound' not in kwargs else kwargs['lbound']
    ubound = 0.04 if 'ubound' not in kwargs else kwargs['ubound']
    cov_windows = 20 if 'cov_windows' not in kwargs else kwargs['cov_windows']
    benchmark_lower = 1.001 if 'benchmark_lower' not in kwargs else kwargs[
        'benchmark_lower']
    benchmark_upper = 0.8 if 'benchmark_upper' not in kwargs else kwargs[
        'benchmark_upper']
    total_lower = -0.001 if 'total_lower' not in kwargs else kwargs[
        'total_lower']
    total_upper = 0.01 if 'total_upper' not in kwargs else kwargs['total_upper']
    effective_industry_lower = 0.0 if 'effective_industry_lower' not in kwargs else kwargs[
        'effective_industry_lower']
    effective_industry_upper = 0.20 if 'effective_industry_upper' not in kwargs else kwargs[
        'effective_industry_upper']
    invalid_industry_lower = 0.0 if 'invalid_industry_lower' not in kwargs else kwargs[
        'invalid_industry_lower']
    invalid_industry_upper = 0.20 if 'invalid_industry_upper' not in kwargs else kwargs[
        'invalid_industry_upper']
    weights_bandwidth = 0.1 if 'weights_bandwidth' not in kwargs else kwargs[
        'weights_bandwidth']
    cov_method = 'unshrunk' if 'cov_method' not in kwargs else kwargs[
        'cov_method']
    method = 'fmv' if 'method' not in kwargs else kwargs['method']
    effective_codes = [] if 'effective_codes' not in kwargs else kwargs[
        'effective_codes']
    invalid_codes = [] if 'invalid_codes' not in kwargs else kwargs[
        'invalid_codes']
    other_boundary = 'absolute' if 'other_boundary' not in kwargs else kwargs[
        'other_boundary']
    benchmark_boundary = 'absolute' if 'benchmark_boundary' not in kwargs else kwargs[
        'benchmark_boundary']
    total_boundary = 'absolute' if 'total_boundary' not in kwargs else kwargs[
        'total_boundary']
    effective_industry_boundary = 'absolute' if 'effective_industry_boundary' not in kwargs else kwargs[
        'effective_industry_boundary']
    invalid_industry_boundary = 'absolute' if 'invalid_industry_boundary' not in kwargs else kwargs[
        'invalid_industry_boundary']
    period = 0 if 'period' not in kwargs else kwargs['period']
    is_benchmark = 0 if 'is_benchmark' not in kwargs else kwargs['is_benchmark']
    neutralized_styles = None if 'neutralized_styles' not in kwargs else kwargs[
        'neutralized_styles']
    synthetize_opts = {} if 'synthetize_opts' not in kwargs else kwargs[
        'synthetize_opts']

    params = {}
    params['industry'] = {}
    params['riskstyle'] = {}

    params['industry']['effective'] = effective_codes
    params['industry']['invalid'] = invalid_codes
    params['riskstyle']['effective'] = []
    params['riskstyle']['invalid'] = []

    params['setting_params'] = {}
    params['setting_params']['period'] = period
    params['setting_params']['weights_bandwidth'] = weights_bandwidth
    params['setting_params']['method'] = method
    params['setting_params']['turn_over_target'] = turn_over_target
    params['setting_params']['target_vol'] = target_vol
    params['setting_params']['cov_windows'] = cov_windows
    params['setting_params']['cov_method'] = cov_method
    params['setting_params']['lbound'] = lbound
    params['setting_params']['ubound'] = ubound
    params['setting_params']['neutralized_styles'] = neutralized_styles
    params['setting_params']['is_benchmark'] = is_benchmark

    params['setting_params']['synthetize_opts'] = synthetize_opts

    params['setting_params']['benchmark'] = {}
    params['setting_params']['total'] = {}
    params['setting_params']['other'] = {}

    ###

    params['setting_params']['other']['boundary'] = other_boundary
    params['setting_params']['other']['lower'] = -0.3
    params['setting_params']['other']['upper'] = 0.3

    # benchmark 区间设置
    params['setting_params']['benchmark']['boundary'] = benchmark_boundary
    params['setting_params']['benchmark']['lower'] = benchmark_lower
    params['setting_params']['benchmark']['upper'] = benchmark_upper

    # total 区间设置   条件6
    params['setting_params']['total']['boundary'] = total_boundary
    params['setting_params']['total']['lower'] = total_lower
    params['setting_params']['total']['upper'] = total_upper

    ### 此处考虑行业择时
    params['setting_params']['effective_industry'] = {}
    params['setting_params']['invalid_industry'] = {}

    #### effective_industry 上限行业区间设置
    params['setting_params']['effective_industry'][
        'boundary'] = effective_industry_boundary
    params['setting_params']['effective_industry'][
        'lower'] = effective_industry_lower
    params['setting_params']['effective_industry'][
        'upper'] = effective_industry_upper

    #### invalid_industry 下限行业区间设置
    params['setting_params']['invalid_industry'][
        'boundary'] = invalid_industry_boundary
    params['setting_params']['invalid_industry'][
        'lower'] = invalid_industry_lower
    params['setting_params']['invalid_industry'][
        'upper'] = invalid_industry_upper
    return params


class RunningSetting(object):

    def __init__(self,
                 lbound=None,
                 ubound=None,
                 weights_bandwidth=None,
                 rebalance_method='risk_neutral',
                 bounds=None,
                 neutralized_styles=None,
                 **kwargs):
        self.lbound = lbound
        self.ubound = ubound
        self.weights_bandwidth = weights_bandwidth
        self.rebalance_method = rebalance_method
        self.bounds = bounds
        self.neutralized_styles = neutralized_styles
        self.more_opts = kwargs


class Strategy(object):

    def __init__(self,
                 alpha_model,
                 features=None,
                 start_date=None,
                 end_date=None,
                 risk_model=None,
                 total_data=None):
        self.start_date = start_date
        self.end_date = end_date
        self.alpha_models = alpha_model
        self.risk_models = risk_model
        self.total_data = total_data
        self.featrues = features
        if self.total_data is not None:
            self.total_data['trade_date'] = pd.to_datetime(
                self.total_data['trade_date'])

    @staticmethod
    def _create_lu_bounds(running_setting, codes, benchmark_w):

        codes = np.array(codes)

        if running_setting.weights_bandwidth:
            lbound = np.maximum(
                0., benchmark_w - running_setting.weights_bandwidth)
            ubound = running_setting.weights_bandwidth + benchmark_w

        lb = running_setting.lbound
        ub = running_setting.ubound

        if lb or ub:
            if not isinstance(lb, dict):
                lbound = np.ones_like(benchmark_w) * lb
            else:
                lbound = np.zeros_like(benchmark_w)
                for c in lb:
                    lbound[codes == c] = lb[c]

                if 'other' in lb:
                    for i, c in enumerate(codes):
                        if c not in lb:
                            lbound[i] = lb['other']
            if not isinstance(ub, dict):
                ubound = np.ones_like(benchmark_w) * ub
            else:
                ubound = np.ones_like(benchmark_w)
                for c in ub:
                    ubound[codes == c] = ub[c]

                if 'other' in ub:
                    for i, c in enumerate(codes):
                        if c not in ub:
                            ubound[i] = ub['other']
        return lbound, ubound

    def _calculate_pos(self,
                       running_setting,
                       er,
                       data,
                       constraints,
                       benchmark_w,
                       lbound,
                       ubound,
                       risk_model,
                       current_position,
                       is_benchmark=0):
        more_opts = running_setting.more_opts
        try:
            target_pos, _ = er_portfolio_analysis(
                er=er,
                industry=data.industry_code.values,
                dx_return=None,
                constraints=constraints,
                detail_analysis=False,
                benchmark=benchmark_w,
                method=running_setting.rebalance_method,
                lbound=lbound,
                ubound=ubound,
                current_position=current_position,
                target_vol=more_opts.get('target_vol'),
                turn_over_target=more_opts.get('turn_over_target'),
                risk_model=risk_model)
        except Exception as e:
            kd_logger.error('{0} rebalance error. {1}'.format(
                data.trade_date.values[0], str(e)))
            target_pos = current_position if not is_benchmark else benchmark_w
            target_pos = pd.DataFrame(target_pos, columns=['weight'])
            target_pos['industry'] = data.industry_code.values
            target_pos['er'] = er
        return target_pos

    def prepare_backtest_models(self, factor_name='factor'):
        models = {}
        total_data_groups = self.total_data.groupby('trade_date')
        alpha_model = ConstLinearModel(features=[factor_name],
                                       weights={factor_name: 1.0})
        for ref_date, _ in total_data_groups:
            models[ref_date] = alpha_model
        self.alpha_models = models
        kd_logger.info("alpha models training finished ...")

    def _build_setting(self, neutralized_styles, effective_industry,
                       invalid_industry, setting_params):
        _boundary = {
            'absolute': BoundaryType.ABSOLUTE,
            'relative': BoundaryType.RELATIVE
        }
        constraint_risk = neutralized_styles
        b_type = []
        l_val = []
        u_val = []
        total_risk_names = constraint_risk + ['benchmark', 'total']
        for name in total_risk_names:
            if name == 'benchmark':
                b_type.append(
                    _boundary[setting_params['benchmark']['boundary']])
                l_val.append(setting_params['benchmark']['lower'])
                u_val.append(setting_params['benchmark']['upper'])
            elif name == 'total':
                b_type.append(_boundary[setting_params['total']['boundary']])
                l_val.append(setting_params['total']['lower'])
                u_val.append(setting_params['total']['upper'])
            elif name in effective_industry:
                b_type.append(_boundary[setting_params['effective_industry']
                                        ['boundary']])
                l_val.append(setting_params['effective_industry']['lower'])
                u_val.append(setting_params['effective_industry']['upper'])
            elif name in invalid_industry:
                b_type.append(
                    _boundary[setting_params['invalid_industry']['boundary']])
                l_val.append(setting_params['invalid_industry']['lower'])
                u_val.append(setting_params['invalid_industry']['upper'])
            else:
                b_type.append(_boundary[setting_params['other']['boundary']])
            l_val.append(setting_params['other']['lower'])
            u_val.append(setting_params['other']['upper'])
        bounds = create_box_bounds(total_risk_names, b_type, l_val, u_val)
        return RunningSetting(
            bounds=bounds,
            lbound=setting_params['lbound'],
            ubound=setting_params['ubound'],
            weights_bandwidth=setting_params['weights_bandwidth'],
            rebalance_method=setting_params['method'],
            cov_winodws=setting_params['cov_windows'],
            cov_method=setting_params['cov_method'],
            target_vol=setting_params['target_vol'],
            is_benchmark=setting_params['is_benchmark'],
            period=setting_params['period'],
            synthetize_opts=setting_params['synthetize_opts'],
            turn_over_target=setting_params['turn_over_target'])

    def _industry_median(self, standard_data, factor_name):
        median_values = standard_data[[
            'trade_date', 'industry_code', 'code', factor_name
        ]].groupby(['trade_date', 'industry_code']).median()[factor_name]
        median_values.name = factor_name + '_median'
        factor_data = standard_data[[
            'trade_date', 'industry_code', 'code', factor_name
        ]].merge(median_values.reset_index(),
                 on=['trade_date', 'industry_code'],
                 how='left')
        factor_data['standard_' + factor_name] = factor_data[factor_name].mask(
            pd.isnull(factor_data[factor_name]),
            factor_data[factor_name + '_median'])
        return factor_data.drop([factor_name + '_median'], axis=1).set_index(
            ['trade_date', 'code', 'industry_code'])

    def _returs_impl(self,
                     price_data,
                     key,
                     name,
                     direction=1):  #1 后置收益率  -1 前置收益率
        price_tb = price_data[key].unstack()
        price_tb.fillna(method='pad', inplace=True)
        return_tb = np.log(price_tb.shift(-1) /
                           price_tb) if direction == 1 else np.log(
                               price_tb / price_tb.shift(1))
        return_tb = return_tb.replace([np.inf, -np.inf], np.nan)
        return_tb = return_tb.stack().reindex(price_data.index)
        return_tb.name = name
        return return_tb

    def create_factors(self, factors_data, params):
        method = params['method']
        if method == 'equal':
            kd_logger.info('synthetize method:{0}'.format(method))
            equal_combine = CombineEngine.create_engine('equal_combine')
            factors_df = factors_data.copy(deep=True)
            factors_df['factor'] = equal_combine(factors_df, self.featrues)
            equal_data = factors_df.copy(deep=True).drop(
                self.featrues, axis=1).sort_values(by=['trade_date', 'code'])
            equal_data['trade_date'] = pd.to_datetime(equal_data['trade_date'])
            return equal_data[['code', 'trade_date', 'factor', 'nxt1_ret']]
        elif method == 'ic_equal':
            kd_logger.info('synthetize method:{0}, span:{1}'.format(
                method, params['span']))
            span = params['span']
            hist_ic_combine = CombineEngine.create_engine('hist_ic_combine')
            factors_df = factors_data.copy(deep=True)
            ic_equal, hist_ret = hist_ic_combine(
                factors_df[['trade_date', 'code'] + self.featrues],
                factors_df[['trade_date', 'code', 'nxt1_ret']],
                self.featrues,
                span=span,
                method='equal')
            ic_equal = ic_equal.rename(columns={
                'combine': 'factor'
            }).merge(factors_df[['trade_date', 'code', 'nxt1_ret']],
                     on=['trade_date',
                         'code']).sort_values(by=['trade_date', 'code'])
            ic_equal['trade_date'] = pd.to_datetime(ic_equal['trade_date'])
            return ic_equal
        elif method == 'ic_half':
            kd_logger.info(
                'synthetize method:{0}, span:{1}, half_life:{2}'.format(
                    method, params['span'], params['half_life']))
            hist_ic_combine = CombineEngine.create_engine('hist_ic_combine')
            span = params['span']
            half_life = params['half_life']
            factors_df = factors_data.copy(deep=True)
            ic_half, hist_ret = hist_ic_combine(
                factors_df[['trade_date', 'code'] + self.featrues],
                factors_df[['trade_date', 'code', 'nxt1_ret']],
                self.featrues,
                span=span,
                method='half_life',
                half_life=half_life)
            ic_half = ic_half.rename(columns={
                'combine': 'factor'
            }).merge(factors_df[['trade_date', 'code', 'nxt1_ret']],
                     on=['trade_date', 'code'
                         ]).sort_values(by=['trade_date', 'code']).sort_values(
                             by=['trade_date', 'code'])
            ic_half['trade_date'] = pd.to_datetime(ic_half['trade_date'])
            return ic_half
        elif method == 'ret_equal':
            kd_logger.info('synthetize method:{0}, span:{1}'.format(
                method, params['span']))
            span = params['span']
            hist_ret_combine = CombineEngine.create_engine('hist_ret_combine')
            factors_df = factors_data.copy(deep=True)
            ret_equal, hist_ret = hist_ret_combine(
                factors_df[['trade_date', 'code'] + self.featrues],
                factors_df[['trade_date', 'code', 'nxt1_ret']],
                self.featrues,
                span=span,
                method='equal')
            ret_equal = ret_equal.rename(columns={
                'combine': 'factor'
            }).merge(factors_df[['trade_date', 'code', 'nxt1_ret']],
                     on=['trade_date',
                         'code']).sort_values(by=['trade_date', 'code'])
            ret_equal['trade_date'] = pd.to_datetime(ret_equal['trade_date'])
            return ret_equal
        elif method == 'ret_half':
            kd_logger.info(
                'synthetize method:{0}, span:{1}, half_life:{2}'.format(
                    method, params['span'], params['half_life']))
            span = params['span']
            half_life = params['half_life']
            hist_ret_combine = CombineEngine.create_engine('hist_ret_combine')
            factors_df = factors_data.copy(deep=True)
            ret_half, hist_ret = hist_ret_combine(
                factors_df[['trade_date', 'code'] + self.featrues],
                factors_df[['trade_date', 'code', 'nxt1_ret']],
                self.featrues,
                span=span,
                method='half_life',
                half_life=half_life)
            ret_half = ret_half.rename(columns={
                'combine': 'factor'
            }).merge(factors_df[['trade_date', 'code', 'nxt1_ret']],
                     on=['trade_date',
                         'code']).sort_values(by=['trade_date', 'code'])
            ret_half['trade_date'] = pd.to_datetime(ret_half['trade_date'])
            return ret_half
        elif method == 'max_ic_sample':
            kd_logger.info(
                'synthetize method:{0}, span:{1}, weight_limit:{2}'.format(
                    method, params['span'], params['weight_limit']))
            max_ic_combine = CombineEngine.create_engine('max_ic_combine')
            span = params['span']
            weight_limit = True if params[
                'weight_limit'] == 1 else False  # 是否约束权重为正
            factors_df = factors_data.copy(deep=True)
            sample_ir_df, hist_ic = max_ic_combine(
                factors_df[['trade_date', 'code'] + self.featrues],
                factors_df[['trade_date', 'code', 'nxt1_ret']],
                self.featrues,
                span=span,
                method='sample',
                weight_limit=weight_limit)
            sample_ir_df = sample_ir_df.rename(columns={
                'combine': 'factor'
            }).merge(factors_df[['trade_date', 'code', 'nxt1_ret']],
                     on=['trade_date',
                         'code']).sort_values(by=['trade_date', 'code'])
            return sample_ir_df
        elif method == 'max_ic_shrunk':
            kd_logger.info(
                'synthetize method:{0}, span:{1}, weight_limit:{2}'.format(
                    method, params['span'], params['weight_limit']))
            max_ic_combine = CombineEngine.create_engine('max_ic_combine')
            span = params['span']
            weight_limit = True if params[
                'weight_limit'] == 1 else False  # 是否约束权重为正
            factors_df = factors_data.copy(deep=True)
            shrunk_ir_df, hist_ic = max_ic_combine(
                factors_df[['trade_date', 'code'] + self.featrues],
                factors_df[['trade_date', 'code', 'nxt1_ret']],
                self.featrues,
                span=span,
                method='shrunk',
                weight_limit=weight_limit)
            shrunk_ir_df = shrunk_ir_df.rename(columns={
                'combine': 'factor'
            }).merge(factors_df[['trade_date', 'code', 'nxt1_ret']],
                     on=['trade_date',
                         'code']).sort_values(by=['trade_date', 'code'])
            return shrunk_ir_df

    def standardize_data(self):
        kd_logger.info("starting standardize data ...")
        kd_logger.info("running setting finished ...")

        factors_data = copy.deepcopy(
            self.total_data)[['trade_date', 'code', 'industry_code'] +
                             self.featrues]
        res = []
        kd_logger.info("starting industry median data ...")
        for col in self.featrues:
            rts = self._industry_median(factors_data, col)
            res.append(rts)
        factors_data = pd.concat(res, axis=1)
        factors_data = factors_data.fillna(0)
        factors_data = factors_data.reset_index().set_index(
            ['trade_date', 'code'])
        industry_dummy = pd.get_dummies(factors_data['industry_code'])
        factors_data = pd.concat([factors_data, industry_dummy], axis=1)
        standarad_cols = ['standard_' + col for col in self.featrues]
        diff_cols = ['trade_date', 'code', 'industry_code']
        grouped = factors_data.reset_index().groupby(['trade_date'])
        kd_logger.info("starting  factor processing data ...")
        alpha_res = []
        industry_styles = factors_data['industry_code'].unique().tolist()
        for k, g in grouped:
            new_factors = factor_processing(
                g[standarad_cols].values,
                pre_process=[winsorize_normal, standardize],
                risk_factors=g[industry_styles].values.astype(float),
                post_process=[standardize])
            f = pd.DataFrame(new_factors, columns=standarad_cols)
            for k in diff_cols:
                f[k] = g[k].values
            alpha_res.append(f)
        factors_data = pd.concat(alpha_res)
        factors_data = factors_data.rename(
            columns=dict(zip(standarad_cols, self.featrues)))
        return factors_data

    def serialize_data(self, factors_data):
        kd_logger.info("running setting finished ...")
        market_data = copy.deepcopy(
            self.total_data)[['trade_date', 'code', 'closePrice']]
        next_rets = self._returs_impl(
            market_data.set_index(['trade_date', 'code']), 'closePrice',
            'nxt1_ret').reset_index()
        next_rets['trade_date'] = pd.to_datetime(next_rets['trade_date'])
        kd_logger.info("returns finished ...")
        serialize_data = copy.deepcopy(factors_data).merge(
            next_rets, on=['trade_date', 'code'])
        return serialize_data

    def synthetize_data(self, serialize_data, params):
        symmetry_othgnz = OthgnzEngine.create_engine('symmetry')
        diff_cols = [
            col for col in serialize_data.columns if col not in self.featrues
        ]
        factors_data = symmetry_othgnz(
            copy.deepcopy(serialize_data[diff_cols + self.featrues]).fillna(0),
            diff_cols).sort_values(by=['trade_date', 'code'])
        running_setting = self._build_setting(
            neutralized_styles=params['industry']['effective'] +
            params['industry']['invalid'],
            effective_industry=params['industry']['effective'],
            invalid_industry=params['industry']['invalid'],
            setting_params=params['setting_params'])
        kd_logger.info("running setting finished ...")
        factors_data = self.create_factors(
            factors_data, running_setting.more_opts['synthetize_opts'])
        return factors_data

    def prepare_backtest_data(self, params):
        standardize_data = self.standardize_data()
        serialize_data = self.serialize_data(standardize_data)
        synthetize_data = self.synthetize_data(serialize_data, params)
        return synthetize_data

    def create_positions(self, params, total_data=None):
        kd_logger.info("starting re-balance ...")
        total_data = total_data if total_data is not None else copy.deepcopy(
            self.total_data)
        is_in_benchmark = (total_data.weight >
                           0.).astype(float).values.reshape((-1, 1))
        total_data.loc[:, 'benchmark'] = is_in_benchmark
        total_data.loc[:, 'total'] = np.ones_like(is_in_benchmark)
        total_data_groups = total_data.groupby('trade_date')
        previous_pos = pd.DataFrame()
        positions = pd.DataFrame()

        running_setting = running_setting = self._build_setting(
            neutralized_styles=params['industry']['effective'] +
            params['industry']['invalid'],
            effective_industry=params['industry']['effective'],
            invalid_industry=params['industry']['invalid'],
            setting_params=params['setting_params'])
        kd_logger.info("running setting finished ...")

        if self.alpha_models is None:
            self.prepare_backtest_models()

        for ref_date, this_data in total_data_groups:
            if ref_date < self.start_date:
                continue
            new_model = self.alpha_models[ref_date]

            risk_model = self.risk_models[
                ref_date] if self.risk_models is not None else None
            codes = this_data.code.values.tolist()

            if previous_pos.empty:
                dates = total_data['trade_date'].dt.strftime(
                    '%Y-%m-%d').unique().tolist()
                try:
                    pos = dates.index(ref_date.strftime('%Y-%m-%d'))
                except:
                    pos = 0
                prev_date = dates[pos]
                current_position = total_data.set_index(
                    'trade_date').loc[prev_date].weight.values
            else:
                previous_pos.set_index('code', inplace=True)
                remained_pos = previous_pos.reindex(codes)
                remained_pos.fillna(0., inplace=True)
                current_position = remained_pos.weight.values
            benchmark_w = this_data.weight.values
            constraints = LinearConstraints(running_setting.bounds, this_data,
                                            benchmark_w)
            lbound, ubound = self._create_lu_bounds(running_setting, codes,
                                                    benchmark_w)
            this_data.fillna(0, inplace=True)
            this_data.fillna(0, inplace=True)
            new_factors = factor_processing(
                this_data[new_model.features].values,
                pre_process=[winsorize_normal, standardize],
                risk_factors=this_data[
                    running_setting.neutralized_styles].values.astype(float)
                if running_setting.neutralized_styles else None,
                post_process=[standardize])
            new_factors = pd.DataFrame(new_factors,
                                       columns=new_model.features,
                                       index=codes)
            er = new_model.predict(new_factors).astype(float)
            kd_logger.info('{0} re-balance: {1} codes'.format(
                ref_date, len(er)))
            target_pos = self._calculate_pos(running_setting,
                                             er,
                                             this_data,
                                             constraints,
                                             benchmark_w,
                                             lbound,
                                             ubound,
                                             risk_model=risk_model,
                                             current_position=current_position)
            target_pos['code'] = codes
            target_pos['trade_date'] = ref_date
            target_pos['benchmark'] = benchmark_w
            positions = positions.append(target_pos)
            previous_pos = target_pos
        return positions

    def rebalance_positions(self, params):
        total_data = self.prepare_backtest_data(params=params)
        industry_code = self.total_data[[
            'trade_date', 'code', 'industry_code'
        ]].set_index(['trade_date', 'code'])
        industry_dummy = pd.get_dummies(industry_code['industry_code'])
        factors_data = pd.concat([
            total_data.set_index(['trade_date', 'code']), industry_dummy,
            industry_code
        ],
                                 axis=1).sort_values(by=['trade_date', 'code'],
                                                     ascending=True)
        index_constituent = self.total_data[['trade_date', 'code',
                                             'weight']].set_index(
                                                 ['trade_date', 'code'])
        factors_data = pd.concat([factors_data, index_constituent], axis=1)
        positions = self.create_positions(
            params=params, total_data=factors_data.reset_index())
        return positions

    def backtest(self, positions, market_data, rate):
        next_rets = self._returs_impl(
            market_data.set_index(['trade_date', 'code']), 'closePrice',
            'nxt1_ret').reset_index()
        kd_logger.info("starting backting ...")
        executor = NaiveExecutor()
        total_data = positions.merge(
            next_rets[['trade_date', 'code', 'nxt1_ret']],
            on=['trade_date', 'code'])
        total_data_groups = total_data.groupby('trade_date')
        rets = []
        b_rets = []
        turn_overs = []
        leverags = []
        for ref_date, this_data in total_data_groups:
            turn_over, executed_pos = executor.execute(this_data)
            leverage = executed_pos.weight.abs().sum()
            ret = executed_pos.weight.values @ (
                np.exp(this_data.nxt1_ret.values) - 1.) - len(
                    this_data.code) * rate * turn_over
            b_ret = executed_pos.benchmark.values @ (
                np.exp(this_data.nxt1_ret.values) - 1.) - len(
                    this_data.code) * rate * turn_over
            rets.append(np.log(1. + ret))
            b_rets.append(np.log(1. + b_ret))
            executor.set_current(executed_pos)
            turn_overs.append(turn_over)
            leverags.append(leverage)
            kd_logger.info(
                '{0}: turn over {1}, returns {2}, benchmark returns {3}'.
                format(ref_date, round(turn_over, 4), round(ret, 4),
                       round(b_ret, 4)))
        trade_dates = positions.trade_date.unique()
        ret_df = pd.DataFrame(
            {
                'returns': rets,
                'turn_over': turn_overs,
                'leverage': leverags,
                'benchmark_returns': b_rets
            },
            index=trade_dates)
        ret_df['excess_return'] = ret_df[
            'returns'] - ret_df['benchmark_returns'] * ret_df['leverage']
        return ret_df

    def matcher(self, ret_df):
        rets = []

        def match(returns, turnover=0, period=empyrical.DAILY):
            returns = returns.astype('float').dropna()
            annual_return = empyrical.annual_return(returns=returns,
                                                    period=period)
            annual_volatility = empyrical.annual_volatility(returns=returns,
                                                            period=period)
            cagr = empyrical.cagr(returns=returns, period=period)
            sharpe_ratio = empyrical.sharpe_ratio(returns=returns,
                                                  period=period)
            downside_risk = empyrical.downside_risk(returns=returns,
                                                    period=period)
            max_drawdown = empyrical.max_drawdown(returns=returns)
            results = {
                'annual_return': annual_return,
                'annual_volatility': annual_volatility,
                'cagr': cagr,
                'sharpe_ratio': sharpe_ratio,
                'downside_risk': downside_risk,
                'max_drawdown': max_drawdown,
                'calmar_ratio': -annual_return / max_drawdown
            }
            if turnover > 0.: results['turnover'] = turnover
            return results

        for r in ['returns', 'benchmark_returns', 'excess_return']:
            turnover = ret_df['turn_over'].mean() if r == 'returns' else 0
            mat = match(ret_df[r], turnover=turnover)
            mat['name'] = r
            rets.append(mat)
        return rets

    def run(self, params, rate=0.0):
        positions = self.rebalance_positions(params)
        market_data = self.total_data[['trade_date', 'code', 'closePrice']]
        ret_df = self.backtest(positions, market_data, rate)
        match_df = self.matcher(ret_df)
        return match_df, ret_df, positions



# -*- coding: utf-8 -*-

import numpy as np
cimport numpy as np
import pandas as pd
cimport cython
from ultron.sentry.Utilities.Tools import to_dict
from ultron.sentry.Analysis.SecurityValueHolders cimport SecurityValueHolder


@cython.boundscheck(False)
@cython.wraparound(False)
cpdef transform(data, list expressions, list cols, str category_field=None, bint to_sort=False, bint dropna=True):

    cdef int dummy_category
    cdef int i
    cdef int j
    cdef np.ndarray total_category
    cdef double[:, :] matrix_values
    cdef double[:, :] output_values
    cdef double[:] narr_view
    cdef size_t start_count
    cdef size_t end_count
    cdef list flags
    cdef list split_category
    cdef list split_values
    cdef SecurityValueHolder exp
    cdef dict dict_data

    if to_sort:
        data.sort_index(inplace=True)

    dummy_category = 0
    if not category_field:
        category_field = 'dummy'
        data[category_field] = 1
        dummy_category = 1
        total_index = list(range(len(data)))
    else:
        total_index = data.index

    total_category = data[category_field].values
    numeric_data = data.select_dtypes([np.number])
    matrix_values = numeric_data.values.astype(float)
    columns = numeric_data.columns.tolist()

    split_category, split_values = to_dict(total_index, total_category.tolist(), matrix_values, columns)

    flags = [isinstance(e, SecurityValueHolder) for e in expressions]
    output_values = np.zeros((len(numeric_data), len(expressions)))

    for i, e in enumerate(expressions):
        if flags[i]:
            if not dummy_category:
                start_count = 0
                for j, dict_data in enumerate(split_values):
                    exp = e
                    exp.push(dict_data)
                    end_count = start_count + len(dict_data)
                    narr_view = exp.value_by_names(split_category[j]).values.astype(float)
                    output_values[start_count:end_count, i] = narr_view
                    start_count = end_count
            else:
                for j, dict_data in enumerate(split_values):
                    exp = e
                    exp.push(dict_data)
                    output_values[j, i] = exp.value_by_name(split_category[j][0])

    df = pd.DataFrame(np.array(output_values), index=data.index, columns=cols)

    for i, e in enumerate(expressions):
        if not flags[i]:
            df[cols[i]] = data[e].values

    if not dummy_category:
        df[category_field] = total_category

    if dropna:
        df.dropna(inplace=True)

    return df



# -*- coding: utf-8 -*-

import unittest
import numpy as np

from ultron.optimize.optimizers import LPOptimizer
from ultron.optimize.optimizers import QuadraticOptimizer
from ultron.optimize.optimizers import TargetVolOptimizer


class Optimizers(unittest.TestCase):
    def test_lpoptimizer(self):
        er = np.array([-1.,-2.])
        lower_bound = np.array([0.,0.2])
        upper_bound = np.array([1.,0.8])

        optimizer = LPOptimizer(objective=-er,
                                cons_matrix=np.array([[1., 1., 1., 1.]]),
                                lbound=lower_bound,
                                ubound=upper_bound)
        self.assertAlmostEqual(optimizer.feval(), 1.2)  
        np.testing.assert_array_almost_equal(optimizer.x_value(), [0.8, 0.2])

    def test_qpoptimizer(self):
        er = np.array([0.01, 0.02, 0.03])
        cov = np.array([[0.02, 0.01, 0.02],
                        [0.01, 0.02, 0.03],
                        [0.02, 0.03, 0.02]])
        ids_var = np.diag([0.01, 0.02, 0.03])
        cov += ids_var
        lbound = np.array([0., 0., 0.])
        ubound = np.array([0.4, 0.4, 0.5])

        cons = np.array([[1., 1., 1.],
                         [1., 0., 1.]])
        clbound = np.array([1., 0.3])
        cubound = np.array([1., 0.7])
        cons_matrix = np.concatenate([cons, clbound.reshape((-1, 1)), cubound.reshape((-1, 1))], axis=1)

        optimizer = QuadraticOptimizer(objective=-er,
                                       cov=cov,
                                       lbound=lbound,
                                       ubound=ubound,
                                       cons_matrix=cons_matrix)
        np.testing.assert_array_almost_equal(optimizer.x_value(),
                                             [0.2, 0.3, 0.5],
                                             4)
    
    def test_qpoptimizer_with_factor_model(self):
        er = np.array([0.1, 0.2, 0.3])
        lbound = np.array([0.0, 0.0, 0.0])
        ubound = np.array([1.0, 1.0, 1.0])

        factor_var = np.array([[0.5, -0.3], [-0.3, 0.7]])
        factor_load = np.array([[0.8, 0.2], [0.5, 0.5], [0.2, 0.8]])
        idsync = np.array([0.1, 0.3, 0.2])

        cons = np.array([[1., 1., 1.]])
        clbound = np.array([1.])
        cubound = np.array([1.])
        cons_matrix = np.concatenate([cons, clbound.reshape((-1, 1)), cubound.reshape((-1, 1))], axis=1)

        optimizer = QuadraticOptimizer(objective=-er,
                                       lbound=lbound,
                                       ubound=ubound,
                                       factor_cov=factor_var,
                                       factor_load=factor_load,
                                       factor_special=idsync,
                                       cons_matrix=cons_matrix)

        # check against cvxpy result
        np.testing.assert_array_almost_equal(optimizer.x_value(),
                                             [0.2866857, 0.21416417, 0.49915014],
                                             4)
    
    def test_qpoptimizer_with_identity_matrix(self):
        er = np.array([-0.02, 0.01, 0.03])
        cov = np.diag([1., 1., 1.])
        optimizer = QuadraticOptimizer(objective=-er,
                                       cov=cov)

        np.testing.assert_array_almost_equal(optimizer.x_value(),
                                             [-0.02, 0.01, 0.03],
                                             4)

    def test_target_vol_optimizer_without_cons(self):
        er = np.array([0.1, 0.2, 0.3])
        cov = np.array([[0.05, 0.01, 0.02],
                        [0.01, 0.06, 0.03],
                        [0.02, 0.03, 0.07]])
        lbound = np.array([-0.3, -0.3, -0.3])
        ubound = np.array([0.5, 0.5, 0.5])
        target_vol = 0.1

        optimizer = TargetVolOptimizer(objective=-er,
                                       cov=cov,
                                       lbound=lbound,
                                       ubound=ubound,
                                       target_vol=target_vol)

        np.testing.assert_array_almost_equal(optimizer.x_value(),
                                             [.0231776, 0.1274768, 0.30130881],
                                             4)

    def test_target_vol_optimizer_with_cons(self):
        er = np.array([0.1, 0.2, 0.3])
        cov = np.array([[0.05, 0.01, 0.02],
                        [0.01, 0.06, 0.03],
                        [0.02, 0.03, 0.07]])
        lbound = np.array([-0.3, -0.3, -0.3])
        ubound = np.array([0.5, 0.5, 0.5])

        cons = np.array([[1., 1., 1.]])
        clbound = np.array([0.])
        cubound = np.array([0.])
        cons_matrix = np.concatenate([cons, clbound.reshape((-1, 1)), cubound.reshape((-1, 1))], axis=1)
        target_vol = 0.1

        optimizer = TargetVolOptimizer(objective=-er,
                                       cov=cov,
                                       lbound=lbound,
                                       ubound=ubound,
                                       target_vol=target_vol,
                                       cons_matrix=cons_matrix)

        np.testing.assert_array_almost_equal(optimizer.x_value(),
                                             [-0.3, -0.10919033, 0.40919033],
                                             4)

    def test_target_vol_optimizer_with_factor_model(self):
        er = np.array([0.1, 0.2, 0.3])
        lbound = np.array([0.0, 0.0, 0.0])
        ubound = np.array([1.0, 1.0, 1.0])

        factor_var = np.array([[0.5, -0.3], [-0.3, 0.7]])
        factor_load = np.array([[0.8, 0.2], [0.5, 0.5], [0.2, 0.8]])
        idsync = np.array([0.1, 0.3, 0.2])

        cons = np.array([[1., 1., 1.]])
        clbound = np.array([1.])
        cubound = np.array([1.])
        target_vol = 0.5
        cons_matrix = np.concatenate([cons, clbound.reshape((-1, 1)), cubound.reshape((-1, 1))], axis=1)

        optimizer = TargetVolOptimizer(objective=-er,
                                       factor_cov=factor_var,
                                       factor_load=factor_load,
                                       factor_special=idsync,
                                       lbound=lbound,
                                       ubound=ubound,
                                       target_vol=target_vol,
                                       cons_matrix=cons_matrix)

        np.testing.assert_array_almost_equal(optimizer.x_value(),
                                             [0.26595552, 0.21675092, 0.51729356],
                                             4)
    def test_target_vol_with_cons_and_ieq(self):
        er = np.array([0.1, 0.2, 0.3])
        cov = np.array([[0.05, 0.01, 0.02],
                        [0.01, 0.06, 0.03],
                        [0.02, 0.03, 0.07]])
        lbound = np.array([-0.3, -0.3, -0.3])
        ubound = np.array([0.5, 0.5, 0.5])

        cons = np.array([[1., 1., 1.]])
        clbound = np.array([0.])
        cubound = np.array([0.])
        target_vol = 0.1

        cons_matrix = np.concatenate([cons, clbound.reshape((-1, 1)), cubound.reshape((-1, 1))], axis=1)

        optimizer = TargetVolOptimizer(objective=-er,
                                       cov=cov,
                                       lbound=lbound,
                                       ubound=ubound,
                                       target_vol=target_vol,
                                       cons_matrix=cons_matrix)

        np.testing.assert_array_almost_equal(optimizer.x_value(),
                                             [-0.3, -0.10919033, 0.40919033],
                                             4)




        




from typing import Iterable
from typing import Dict
import pandas as pd
import numpy as np
import pickle, time, itertools, datetime, hashlib
from copy import copy
from joblib import Parallel, delayed
from ultron.factor.genetic.geneticist.operators import expression_params, calc_factor
from ultron.factor.fitness.weighted import Weighted
from ultron.utilities.utils import check_random_state
from ultron.utilities.jobs import partition_estimators
from ultron.utilities.utils import check_random_state

import warnings

warnings.filterwarnings("ignore")

MAX_INT = np.iinfo(np.int32).max
MIN_INT = np.iinfo(np.int32).min


class Individual(object):

    def __init__(self, name, stype):  # 0 行业  1 股票池
        self._name = name
        self._stype = stype


class Program(object):

    def __init__(self,
                 gen,
                 random_state,
                 ind_sets,
                 universe_sets,
                 init_depth,
                 p_point_replace,
                 expression,
                 program=None,
                 parents=None):
        self._gen = gen
        self._p_point_replace = p_point_replace
        self._ind_sets = ind_sets
        self._universe_sets = universe_sets
        self._program = program
        self._parents = parents
        self._ind_sets = ind_sets
        self._universe_sets = universe_sets
        self._create_time = datetime.datetime.now()
        self._init_depth = init_depth
        self._expression = expression
        self._raw_fitness = 0
        self._is_valid = True
        self._risk_styles = [
            'BETA', 'MOMENTUM', 'SIZE', 'EARNYILD', 'RESVOL', 'GROWTH', 'BTOP',
            'LEVERAGE', 'LIQUIDTY', 'SIZENL'
        ]
        self._industry_styles = [
            'Bank', 'RealEstate', 'Health', 'Transportation', 'Mining',
            'NonFerMetal', 'HouseApp', 'LeiService', 'MachiEquip', 'BuildDeco',
            'CommeTrade', 'CONMAT', 'Auto', 'Textile', 'FoodBever',
            'Electronics', 'Computer', 'LightIndus', 'Utilities', 'Telecom',
            'AgriForest', 'CHEM', 'Media', 'IronSteel', 'NonBankFinan',
            'ELECEQP', 'AERODEF', 'Conglomerates'
        ]
        self._name = 'ultron_' + str(
            int(time.time() * 1000000 + datetime.datetime.now().microsecond))
        if self._program is None:
            self._program = self.build_program(random_state)
        self.create_identification()

    def create_identification(self):
        m = hashlib.md5()
        program = [p._name for p in self._program]
        program.sort()
        m.update(bytes(''.join(program), encoding='UTF-8'))
        self._identification = m.hexdigest()

    def industry_features(self):
        return [
            p._name for p in self._program if p._name in self._industry_styles
        ]

    def log(self):
        parents = {'method': 'gen'} if self._parents is None else self._parents
        name = self._name
        sets = [p._name for p in self._program]
        print(
            'name:%s,raw_fitness:%f, program:%s,create_time:%s,identification:%s'
            % (self._name, self._raw_fitness, sets, self._create_time,
               self._identification))

    def duplicate_removal(self, program):
        program = [p._name for p in program]
        program = set(program)
        new_program = []
        for p in program:
            if p in self._ind_sets:
                new_program.append(Individual(p, 0))
            elif p in self._universe_sets:
                new_program.append(Individual(p, 1))
        return new_program

    def build_program(self, random_state):
        #构造 深度判断
        ind_max = self._init_depth[1] if self._init_depth[1] < len(
            self._ind_sets) else len(self._ind_sets)
        universe_max = self._init_depth[1] if self._init_depth[1] < len(
            self._universe_sets) else len(self._universe_sets)
        ind_max_prob = np.random.randint(ind_max) / len(self._ind_sets)
        universe_max_prob = np.random.randint(universe_max) / len(
            self._universe_sets)
        univ = self._universe_sets[np.random.randint(len(self._universe_sets))]
        ind = self._ind_sets[np.random.randint(len(self._ind_sets))]
        program = [Individual(univ, 1), Individual(ind, 0)]
        ind_list = (np.random.uniform(0, 1, (1, len(self._ind_sets))) <
                    ind_max_prob).tolist()
        ind_list = list(np.array(self._ind_sets)[ind_list])
        universe_list = (np.random.uniform(0, 1, (1, len(self._universe_sets)))
                         < universe_max_prob).tolist()
        universe_list = list(
            set(list(np.array(self._universe_sets)[universe_list])))
        program += [Individual(ind, 0) for ind in ind_list
                    ] + [Individual(univ, 1) for univ in universe_list]
        return self.duplicate_removal(program)

    def get_subtree(self, random_state=None, program=None):
        if program is None:
            program = self._program
        probs = np.array(
            [0.9 if node._name in self._ind_sets else 0.1 for node in program])
        probs = np.cumsum(probs / probs.sum())
        start = np.searchsorted(probs, np.random.uniform())
        end = start if (len(program) - 1) <= start else np.random.randint(
            start,
            len(program) - 1)
        return start, end

    def reproduce(self):
        return copy(self._program)

    def crossover(self, donor, random_state):
        start, end = self.get_subtree(random_state)
        end -= 1
        removed = range(start, end)
        donor_start, donor_end = self.get_subtree(random_state, donor)
        donor_removed = list(
            set(range(len(donor))) - set(range(donor_start, donor_end)))
        return self.duplicate_removal(
            self._program[:start] + donor[donor_start:donor_end] +
            self._program[end:]), removed, donor_removed

    ##树变异
    def subtree_mutation(self, random_state=None):
        chicken = self.build_program(random_state)
        return self.crossover(chicken, random_state)

    ##突变异
    def hoist_mutation(self, random_state):
        start, end = self.get_subtree(random_state)
        subtree = self._program[start:end]
        sub_start, sub_end = self.get_subtree(random_state, subtree)
        hoist = subtree[sub_start:sub_end]
        removed = list(
            set(range(start, end)) -
            set(range(start + sub_start, start + sub_end)))
        return self.duplicate_removal(self._program[:start] + hoist +
                                      self._program[end:]), removed

    def point_mutation(self, random_state):
        program = copy(self._program)
        mutate = np.where(
            np.random.uniform(size=len(program)) < self._p_point_replace)[0]
        for node in mutate:
            if program[node]._name in self._ind_sets:
                replace_node = self._ind_sets[np.random.randint(
                    0,
                    len(self._ind_sets) - 1)]
                program[node] = Individual(replace_node, 0)
            elif program[node]._name in self._universe_sets:
                replace_node = self._universe_sets[np.random.randint(
                    0,
                    len(self._universe_sets) - 1)]
                program[node] = Individual(replace_node, 1)
        return self.duplicate_removal(program), list(mutate)

    def _industry_expression(self, industry_styles, name):
        expression = ''
        industry_len = len(industry_styles)
        i = 1
        for industry in industry_styles:
            expression += """({0}[\'{1}\']==1)""".format(name, industry)
            if i < (industry_len):
                expression += """|"""
            i += 1
        return expression

    def optimization_by_formula(self, standard_data, univ_df, expression):
        #print('program:', self._program, isinstance(self._program, list))
        if not isinstance(self._program, list):
            pass
        universe_codes = [
            p._name for p in self._program if p._name in self._universe_sets
        ]
        industry_styles = [
            p._name for p in self._program if p._name in self._ind_sets
        ]
        univ_se = univ_df.set_index(['trade_date','code'])[universe_codes] if len(universe_codes) > 0 else \
                        univ_df.set_index(['trade_date','code'])
        univ_se = univ_se[univ_se > 0]  # 股票池
        standard_factors = standard_data.copy()
        standard_datadard_factors = standard_factors.set_index(
            ['trade_date', 'code']).reindex(univ_se.index).reset_index()
        if len(industry_styles) > 0:
            filter_expression = self._industry_expression(
                industry_styles, 'standard_factors')
            standard_factors = standard_factors[eval(filter_expression)]
        return calc_factor(expression=expression,
                           total_data=standard_factors,
                           indexs='trade_date',
                           key='code',
                           name='transformed').reset_index().dropna()

    def evaluation(self,
                   standard_data,
                   univ_df,
                   default_value,
                   expression,
                   fitness_method='high_frequency',
                   horizon=2,
                   is_neutralize=0):
        risk_data = standard_data[['trade_date', 'code'] +
                                  self._industry_styles + self._risk_styles]
        mkt_df = standard_data[[
            'closePrice', 'openPrice', 'lowestPrice', 'highestPrice',
            'turnoverVol', 'turnoverValue', 'trade_date', 'code',
            'accumAdjFactor', 'chgPct', 'bar30_vwap', 'bar60_vwap'
        ]]
        factors_data = self.optimization_by_formula(standard_data, univ_df,
                                                    expression)
        weight = Weighted.create_weighted(method=fitness_method)
        result = weight.run(factor_data=factors_data,
                            horizon=horizon,
                            risk_data=risk_data,
                            mkt_df=mkt_df,
                            factor_name='transformed',
                            is_neutralize=is_neutralize,
                            default_value=np.iinfo(np.int32))
        raw_fitness = default_value if result['status'] == 0 else abs(
            result['score'])
        self._raw_fitness = default_value if np.isnan(
            raw_fitness) else raw_fitness
        self._is_valid = False if self._raw_fitness == default_value else True


def parallel_evolve(n_programs, parents, standard_data, univ_df, seeds,
                    ind_sets, universe_sets, gen, params):
    tournament_size = params['tournament_size']
    init_depth = params['init_depth']
    greater_is_better = params['greater_is_better']
    expression = params['expression']
    method_probs = params['method_probs']
    p_point_replace = params['p_point_replace']
    expression = params['expression']

    programs = []

    def _contenders(tour_parents):
        contenders = random_state.randint(0, len(tour_parents), 2)
        return [tour_parents[p] for p in contenders]

    def _tournament(tour_parents):
        contenders = random_state.randint(0, len(tour_parents),
                                          tournament_size)
        raw_fitness = [tour_parents[p]._raw_fitness for p in contenders]
        if greater_is_better:
            parent_index = contenders[np.argmax(raw_fitness)]
        else:
            parent_index = contenders[np.argmin(raw_fitness)]
        return tour_parents[parent_index], parent_index

    for i in range(n_programs):
        random_state = check_random_state(seeds[i])
        if parents is None:
            program = None
            genome = None
        else:
            method = random_state.uniform()
            parent, parent_index = _tournament(copy(parents))
            ori_parent = copy(parent)
            contenders = _contenders(copy(parents))
            for contender in contenders:
                program, removed, remains = parent.crossover(
                    contender._program, random_state)
                parent = Program(gen=gen,
                                 random_state=random_state,
                                 ind_sets=ind_sets,
                                 universe_sets=universe_sets,
                                 init_depth=init_depth,
                                 p_point_replace=p_point_replace,
                                 expression=expression,
                                 program=program,
                                 parents=parents)

            #新特征种群加入
            if random_state.uniform() < method_probs[2]:
                program = Program(gen=gen,
                                  random_state=random_state,
                                  ind_sets=ind_sets,
                                  universe_sets=universe_sets,
                                  init_depth=init_depth,
                                  p_point_replace=p_point_replace,
                                  expression=expression,
                                  program=None)
                program, removed, remains = parent.crossover(
                    program._program, random_state)
                parent = Program(gen=gen,
                                 random_state=random_state,
                                 ind_sets=ind_sets,
                                 universe_sets=universe_sets,
                                 init_depth=init_depth,
                                 p_point_replace=p_point_replace,
                                 expression=expression,
                                 program=program)

            if method < method_probs[0]:  # # crossover
                donor, donor_index = _tournament(copy(parents))
                program, removed, remains = parent.crossover(
                    donor._program, random_state)
                genome = {
                    'method': 'Crossover',
                    'parent_idx': parent_index,
                    'parent_nodes': removed,
                    'donor_idx': donor_index,
                    'donor_nodes': remains
                }
            elif method < method_probs[1]:  # subtree_mutation
                program, removed, _ = parent.subtree_mutation(random_state)
                genome = {
                    'method': 'Subtree Mutation',
                    'parent_idx': parent_index,
                    'parent_nodes': removed
                }
            elif method < method_probs[2]:  # hoist_mutation
                program, removed = parent.hoist_mutation(random_state)
                genome = {
                    'method': 'Hoist Mutation',
                    'parent_idx': parent_index,
                    'parent_nodes': removed
                }
            elif method < method_probs[3]:  # point_mutation
                program, mutated = parent.point_mutation(random_state)
                genome = {
                    'method': 'Point Mutation',
                    'parent_idx': parent_index,
                    'parent_nodes': mutated
                }
            else:
                program = parent.reproduce()  # reproduction
                genome = {
                    'method': 'Reproduction',
                    'parent_idx': parent_index,
                    'parent_nodes': []
                }

            # 与原始自身进行交叉
            if random_state.uniform() < method_probs[3]:
                program = Program(gen=gen,
                                  random_state=random_state,
                                  ind_sets=ind_sets,
                                  universe_sets=universe_sets,
                                  init_depth=init_depth,
                                  p_point_replace=p_point_replace,
                                  expression=expression,
                                  program=program,
                                  parents=genome)
                program, removed, remains = program.crossover(
                    ori_parent._program, random_state)

        program = Program(gen=gen,
                          random_state=random_state,
                          ind_sets=ind_sets,
                          universe_sets=universe_sets,
                          init_depth=init_depth,
                          p_point_replace=p_point_replace,
                          expression=expression,
                          program=program,
                          parents=genome)
        default_value = MIN_INT if greater_is_better else MAX_INT
        program.evaluation(standard_data, univ_df, default_value, expression)
        programs.append(program)
    return programs


class Gentic(object):

    def __init__(self,
                 ind_sets,
                 universe_sets,
                 random_state=None,
                 population_size=5,
                 tournament_size=2,
                 generations=MAX_INT,
                 greater_is_better=True,
                 verbose=1,
                 convergence=0.05,
                 init_depth=(5, 6),
                 n_jobs=1,
                 p_crossover=0.9,
                 p_subtree_mutation=0.01,
                 p_hoist_mutation=0.01,
                 p_point_mutation=0.01,
                 p_point_replace=0.05):
        self._random_state = random_state
        self._generations = MAX_INT if generations == 0 else generations
        self._population_size = population_size
        self._tournament_size = tournament_size
        self._init_depth = init_depth
        self._n_jobs = n_jobs
        self._greater_is_better = greater_is_better
        self._p_crossover = p_crossover
        self._p_subtree_mutation = p_subtree_mutation
        self._p_hoist_mutation = p_hoist_mutation
        self._p_point_mutation = p_point_mutation
        self._p_point_replace = p_point_replace
        self._convergence = convergence
        self._verbose = verbose
        self._ind_sets = ind_sets
        self._universe_sets = universe_sets

    def train(self, standard_data, univ_df, expression):
        random_state = check_random_state(self._random_state)
        self._method_probs = np.array([
            self._p_crossover, self._p_subtree_mutation,
            self._p_hoist_mutation, self._p_point_mutation
        ])

        self._method_probs = np.cumsum(self._method_probs)

        if self._method_probs[-1] > 1:
            raise ValueError('The sum of p_crossover, p_subtree_mutation, '
                             'p_hoist_mutation and p_point_mutation should '
                             'total to 1.0 or less.')

        if (isinstance(self._init_depth, tuple)
                and len(self._init_depth) != 2):
            raise ValueError('init_depth should be a tuple with length two.')

        if (isinstance(self._init_depth, tuple)
                and (self._init_depth[0] > self._init_depth[1])):
            raise ValueError('init_depth should be in increasing numerical '
                             'order: (min_depth, max_depth).')

        params = {}
        params['tournament_size'] = self._tournament_size
        params['p_point_replace'] = self._p_point_replace
        params['init_depth'] = self._init_depth
        params['greater_is_better'] = self._greater_is_better
        params['expression'] = expression
        params['method_probs'] = self._method_probs
        params['p_point_replace'] = self._p_point_replace

        count_time = 0
        iteration_best_programs = None
        self._programs = []
        run_details = {
            'generation': [],
            'average_fitness': [],
            'best_fitness': [],
            'generation_time': [],
            'best_programs': []
        }

        prior_generations = len(self._programs)
        n_more_generations = self._generations - prior_generations
        for gen in range(prior_generations, self._generations):
            start_time = time.time()
            if gen == 0:
                parents = None
            else:
                parents = self._programs[gen - 1]

            n_jobs, n_programs, starts = partition_estimators(
                self._population_size, self._n_jobs)

            seeds = random_state.randint(MAX_INT, size=self._population_size)
            population = Parallel(n_jobs=n_jobs, verbose=self._verbose)(
                delayed(parallel_evolve)(n_programs[i], parents, standard_data,
                                         univ_df, seeds, self._ind_sets,
                                         self._universe_sets, gen, params)
                for i in range(n_jobs))
            population = list(itertools.chain.from_iterable(population))
            #有效性判断
            population = [p for p in population if p._is_valid]

            self._programs.append(population)

            if iteration_best_programs is not None:
                programs = np.concatenate(
                    [population, iteration_best_programs])
            else:
                programs = population

            fitness = [program._raw_fitness for program in programs]
            if self._greater_is_better:
                iteration_best_programs = np.array(programs)[np.argsort(
                    fitness)[-self._tournament_size:]]
            else:
                iteration_best_programs = np.array(programs)[np.argsort(
                    fitness)[:self._tournament_size:]]

            fitness = [
                program._raw_fitness for program in iteration_best_programs
            ]

            for program in iteration_best_programs:
                program.log()

            run_details['generation'].append(gen)
            run_details['average_fitness'].append(np.mean(fitness))
            generation_time = time.time() - start_time
            run_details['generation_time'].append(generation_time)
            run_details['best_programs'].append(iteration_best_programs)
            print(np.mean(fitness))

            if gen == 0:
                continue

            d_value = np.mean(fitness) - run_details['average_fitness'][gen -
                                                                        1]
            print('d_value:%f,convergence:%f,count:%d,con_time:%d' %
                  (d_value, self._convergence, len(iteration_best_programs),
                   count_time))
            if abs(d_value) < self._convergence:
                count_time += 1
                if count_time > 3:
                    break
            else:
                count_time = 0

        return run_details


# -*- coding: utf-8 -*-
import pandas as pd
from . cov_method import (
    unshrunk_cov, lwshrunk_cov, shrunk_cov, gridsearch_cov, oas_cov)

func_dict = {'unshrunk':unshrunk_cov, 'lwshrunk':lwshrunk_cov,
            'shrunk':shrunk_cov, 'gridsearch':gridsearch_cov, 
            'oas':oas_cov}

class CovEngine(object):
    @classmethod
    def create_engine(cls, ce_name):
        return func_dict[ce_name]

    @classmethod
    def calc_cov(cls, name, ret_tb, window=20, is_pandas=True):
        cov = func_dict[name](ret_tb=ret_tb, window=window)
        if not is_pandas:
            return cov
        codes = ret_tb.columns.get_level_values(1).unique()
        return pd.DataFrame(data=cov,index=codes,columns=codes)


import sys, os
from IPython.display import clear_output
from IPython.display import display
from ipywidgets import FloatProgress

from ultron.kdutils.decorator import warnings_filter


class Progress(object):
    """单进程（主进程）进度显示控制类"""

    @warnings_filter
    def __init__(self, total, a_progress, label=None):
        """
        外部使用eg：
            progess = Progress(stock_df.shape[0], 0, 'merging {}'.format(m))
            for i, symbol in enumerate(stock_df['symbol']):
                progess.show(i + 1)
        :param total: 总任务数量
        :param a_progress: 初始进度
        :param label: 进度显示label
        """
        self._total = total
        self._progress = a_progress
        self._label = label
        self.f = sys.stdout
        self.progress_widget = None

    def __enter__(self):
        """创建子进程做进度显示"""
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.f.write('\r')
        if self.progress_widget is not None:
            self.progress_widget.close()

    @property
    def progress(self):
        """property获取self._progress"""
        return self._progress

    @progress.setter
    def progress(self, a_progress):
        """progress.setter设置progress"""
        if a_progress > self._total:
            self._progress = self._total
        elif a_progress < 0:
            self._progress = 0
        else:
            self._progress = a_progress

    def show(self, a_progress=None, ext='', p_format="{}:{}:{}%"):
        """
        进行进度控制显示主方法
        :param ext: 可以添加额外的显示文字，str，默认空字符串
        :param a_progress: 默认None, 即使用类内部计算的迭代次数进行进度显示
        :param p_format: 进度显示格式，默认{}: {}%，即'self._label:round(self._progress / self._total * 100, 2))%'
        """
        self.progress = a_progress if a_progress is not None else self.progress + 1
        ps = round(self._progress / self._total * 100, 2)

        if self._label is not None:
            # 如果初始化label没有就只显示ui进度
            self.f.write('\r')
            self.f.write(p_format.format(self._label, ext, ps))

        if 'IS_IPYTHON' in os.environ and os.environ['IS_IPYTHON']:
            if self.progress_widget is None:
                self.progress_widget = FloatProgress(value=0, min=0, max=100)
                display(self.progress_widget)
            self.progress_widget.value = ps

        # 这样会出现余数结束的情况，还是尽量使用上下文管理器控制结束
        if self._progress == self._total:
            self.f.write('\r')
            if self.progress_widget is not None:
                self.progress_widget.close()



# -*- encoding:utf-8 -*-
"""示例ump边裁特征价格模块"""

from ultron.ump.trade.ml_feature import FeaturePrice
from ultron.ump.ump.base import ump_edge_make_xy, BuyUmpMixin
from ultron.ump.ump.edge_base import EdgeBase
from ultron.ump.model.principles import Principles


class PriceFiter(Principles):
    """
        内部类，只需要实现make_xy，且使用ump_edge_make_xy装饰

        在边裁__init__中通过：
                self.fiter_cls = self.get_fiter_class()
                self.fiter = self.fiter_cls(orders_pd=orders_pd, **kwarg)
        构造裁判的filter以及重要的self.fiter.df，即pd.DataFrame对象特征
    """

    @ump_edge_make_xy
    def make_xy(self, **kwarg):
        """
         make_xy通过装饰器ump_edge_make_xy进行包装，调用make_xy前将有交易结果的单子进行筛选：
         order_has_ret = orders_pd[orders_pd['result'] != 0]，且赋予self.order_has_ret
         make_xy只需要使用filter选取需要的特征，即从self.order_has_ret中使用filter选取特征列形成df

        :param kwarg: ump_edge_make_xy装饰器中使用kwarg
                      kwargs['orders_pd'] 做为必须要有的关键字参数：交易训练集数据，pd.DataFrame对象
        :return: self.order_has_ret中使用filter选取特征列形成price_df
                 ump_edge_make_xy装饰器在make_xy返回price_df后做转换matrix，形成x，y等工作
        """

        filter_list = ['profit', 'profit_cg']
        cols = FeaturePrice().get_feature_ump_keys(ump_cls=EdgePrice)
        # ['profit', 'profit_cg', 'buy_price_rank120', 'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252']
        filter_list.extend(cols)

        # noinspection PyUnresolvedReferences
        price_df = self.order_has_ret.filter(filter_list)
        """
            eg: price_df
                           profit  profit_cg  buy_price_rank120  buy_price_rank90  \
            2014-09-24  -22618.04    -0.0566              1.000             1.000
            2014-10-24  -29690.28    -0.0742              1.000             1.000
            2014-10-29   18959.19     0.0542              1.000             1.000
            2014-10-29  148209.36     0.5022              0.925             0.900
            2014-10-29   24867.60     0.0952              0.962             0.950
            2014-10-29   18959.19     0.0542              1.000             1.000
            2014-11-03    1250.80     0.0045              1.000             1.000
            2014-11-11   59888.21     0.1857              0.954             0.939
            2014-11-12   -3578.78    -0.0140              0.475             0.522
            2014-11-26  -29085.19    -0.0946              0.642             0.733
            ...               ...        ...                ...               ...
            2016-03-14   16220.57     0.0559              0.617             0.500
            2016-03-14  -25328.12    -0.1218              0.683             0.589
            2016-03-30  -29858.44    -0.0863              0.658             0.667
            2016-04-04    5373.76     0.0244              0.400             0.511
            2016-04-13  -28044.40    -0.1159              0.567             0.722
            2016-04-14  -18645.93    -0.0467              0.875             0.878
            2016-04-15  -32484.79    -0.1149              0.775             0.733
            2016-04-15  -32484.79    -0.1149              0.775             0.733
            2016-04-29     290.96     0.0007              1.000             1.000
            2016-04-29     290.96     0.0007              1.000             1.000

                        buy_price_rank60  buy_price_rank252
            2014-09-24             1.000              1.000
            2014-10-24             1.000              1.000
            2014-10-29             1.000              1.000
            2014-10-29             0.883              0.750
            2014-10-29             0.925              0.982
            2014-10-29             1.000              1.000
            2014-11-03             1.000              1.000
            2014-11-11             0.992              0.808
            2014-11-12             0.783              0.560
            2014-11-26             1.000              0.762
            ...                      ...                ...
            2016-03-14             0.750              0.444
            2016-03-14             0.850              0.623
            2016-03-30             1.000              0.536
            2016-04-04             0.767              0.190
            2016-04-13             1.000              0.270
            2016-04-14             0.967              0.940
            2016-04-15             1.000              0.631
            2016-04-15             1.000              0.631
            2016-04-29             1.000              1.000
            2016-04-29             1.000              1.000
        """

        return price_df


class EdgePrice(EdgeBase, BuyUmpMixin):
    """边裁价格特征类，UltronUmpEdgeBase子类，混入BuyUmpMixin，做为买入ump类"""

    def get_predict_col(self):
        """
        边裁价格特征keys：['buy_price_rank120', 'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252']
        :return: ['buy_price_rank120', 'buy_price_rank90', 'buy_price_rank60', 'buy_price_rank252']
        """
        return FeaturePrice().get_feature_ump_keys(ump_cls=EdgePrice)

    def get_fiter_class(self):
        """
        边裁价格特征返回的UltronMLPd子类：UltronUmpEdgePrice.UmpPriceFiter
        :return: UltronUmpEdgePrice.UmpPriceFiter
        """
        return PriceFiter

    @classmethod
    def class_unique_id(cls):
        """
         具体ump类关键字唯一名称，类方法：return 'price_edge'
         主要针对外部user设置自定义ump使用, 需要user自己保证class_unique_id的唯一性，内部不做检测
         具体使用见UltronUmpManager中extend_ump_block方法
        """
        return 'price_edge'